# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Disk backed cache for preserving values across runs of
run.py."""

import hashlib
import os

from clusterfuzz._internal.base import json_utils
from clusterfuzz._internal.metrics import logs
from clusterfuzz._internal.system import environment

# For any given value file, if a file with the same name with this added
# extension exists, it is not cleared during initialization.
PERSIST_FILE_EXTENSION = '.persist'


def initialize():
  """Initialize the persistent cache, creating the directory used to store the
  values."""
  cache_directory_path = environment.get_value('CACHE_DIR')
  if os.path.exists(cache_directory_path):
    clear_values()
  else:
    os.makedirs(cache_directory_path)


def clear_values(clear_all=False):
  """Remove all values."""
  cache_directory_path = environment.get_value('CACHE_DIR')
  if not os.path.exists(cache_directory_path):
    return

  for root_directory, _, filenames in os.walk(cache_directory_path):
    for filename in filenames:
      if filename.endswith(PERSIST_FILE_EXTENSION) and not clear_all:
        continue

      file_path = os.path.join(root_directory, filename)
      persist_file_path = file_path + PERSIST_FILE_EXTENSION
      if os.path.exists(persist_file_path) and not clear_all:
        continue

      os.remove(file_path)


def delete_value(key):
  """Removes the value for a key."""
  value_path = get_value_file_path(key)
  if os.path.exists(value_path):
    os.remove(value_path)


def get_value(key, default_value=None, constructor=None):
  """Get the value for a key."""
  value_path = get_value_file_path(key)

  if not os.path.exists(value_path):
    return default_value

  try:
    with open(value_path, 'rb') as f:
      value_str = f.read()
  except IOError:
    logs.log_error('Failed to read %s from persistent cache.' % key)
    return default_value

  try:
    value = json_utils.loads(value_str)
  except Exception:
    logs.log_warn('Non-serializable value read from cache key %s: "%s"' %
                  (key, value_str))
    return default_value

  if constructor:
    try:
      value = constructor(value)
    except Exception:
      logs.log_warn('Failed to construct value "%s" using %s '
                    'and key "%s" in persistent cache. Using default value %s.'
                    % (value, constructor, key, default_value))
      return default_value

  return value


def get_value_file_path(key):
  """Return the full path to the value file for the given key."""
  # Not using utils.string_hash here to avoid a circular dependency.
  # TODO(mbarbella): Avoid this once utils.py is broken into multiple files.
  key_filename = 'cache-%s.json' % hashlib.sha1(str(key).encode()).hexdigest()
  cache_directory_path = environment.get_value('CACHE_DIR')
  return os.path.join(cache_directory_path, key_filename)


def set_value(key, value, persist_across_reboots=False):
  """Set the value for a key. If |persist_across_restarts| is set, then the key
  won't be deleted even run.py is restarted. """
  value_path = get_value_file_path(key)

  try:
    value_str = json_utils.dumps(value)
  except Exception:
    logs.log_error(
        'Non-serializable value stored to cache key %s: "%s"' % (key, value))
    return

  try:
    with open(value_path, 'wb') as f:
      f.write(value_str.encode())
  except IOError:
    logs.log_error('Failed to write %s to persistent cache.' % key)

  if not persist_across_reboots:
    return

  persist_value_path = value_path + PERSIST_FILE_EXTENSION
  if os.path.exists(persist_value_path):
    return

  try:
    open(persist_value_path, 'wb').close()
  except IOError:
    logs.log_error(
        'Failed to write presistent metadata file for cache key %s' % key)



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Base Class for the Delta and Chunk minimizers."""

import unittest


class BaseMinimizerTester(unittest.TestCase):
  """Base class for chunk and delta minimizer tests."""

  def setUp(self):
    raise NotImplementedError

  def _mock_test_function(self, data_file):
    """Mock test function to reduce time minimizer takes and simplify tests."""
    data = open(data_file, 'rb').read()
    if b'error' in data:
      return False
    return True

  def test_minimizer_still_contains_error(self):
    """Tests a simple minimization. Should remove the first and last lines."""
    testcase = b'x = 2 \n error \n return x'
    minimized = self.line_minimizer.minimize(testcase)
    self.assertEqual(minimized, b' error ')

  def test_minimizer_accepts_empty_data(self):
    """Tests that the minimizer does not break on empty string."""
    testcase = b''
    minimized = self.line_minimizer.minimize(testcase)
    self.assertEqual(minimized, b'')

  def test_minimizer_does_not_minimize_non_errored_code(self):
    """"Test that the minimizer does not break when there is nothing to min."""
    testcase = b'This \n Code \n Has \n No \n Error'
    minimized = self.line_minimizer.minimize(testcase)
    self.assertEqual(minimized, testcase)



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for app functions."""

import os

from clusterfuzz._internal.platforms.android import app
from clusterfuzz._internal.system import environment
from clusterfuzz._internal.tests.test_libs import android_helpers


class IsInstalledTest(android_helpers.AndroidTest):
  """Tests is_installed."""

  def test_nonexistent_package_not_installed(self):
    """Ensure that a non-existent package is not installed."""
    self.assertFalse(app.is_installed('non.existent.package'))

  def test_partial_package_name_not_installed(self):
    """Test that com.google is not recognized as an installed package."""
    self.assertFalse(app.is_installed('com.google'))

  def test_package_installed(self):
    """Ensure that gms (which should always be available) is installed."""
    self.assertTrue(app.is_installed('com.google.android.gms'))


class GetPackageNameTest(android_helpers.AndroidTest):
  """Tests get_package_name."""

  def setUp(self):
    super(GetPackageNameTest, self).setUp()

    root_dir = environment.get_value('ROOT_DIR')
    self.test_apk_path = os.path.join(root_dir, 'resources', 'platform',
                                      'android', 'wifi_util.apk')
    self.test_apk_pkg_name = 'com.android.tradefed.utils.wifi'

  def test_pkg_name_in_env(self):
    """Test package name already set in |PKG_NAME| env."""
    environment.set_value('PKG_NAME', 'a.b.c')
    self.assertEqual(app.get_package_name(), 'a.b.c')

  def test_apk_path_in_app_path_env(self):
    """Test apk path set in |APP_PATH| env variable."""
    environment.set_value('APP_PATH', self.test_apk_path)
    self.assertEqual(app.get_package_name(), self.test_apk_pkg_name)

  def test_apk_path_in_arg(self):
    """Test apk path passed as argument."""
    self.assertEqual(
        app.get_package_name(self.test_apk_path), self.test_apk_pkg_name)



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Configuration Manager."""

from flask import request

from clusterfuzz._internal.base import utils
from clusterfuzz._internal.config import db_config
from clusterfuzz._internal.datastore import data_types
from handlers import base_handler
from libs import form
from libs import handler
from libs import helpers

USER_PERMISSION_ENTITY_KINDS = [
    {
        'name': 'fuzzer',
        'value': data_types.PermissionEntityKind.FUZZER,
    },
    {
        'name': 'job',
        'value': data_types.PermissionEntityKind.JOB,
    },
    {
        'name': 'uploader',
        'value': data_types.PermissionEntityKind.UPLOADER,
    },
]

USER_PERMISSION_AUTO_CC_TYPES = [
    {
        'name': 'none',
        'value': data_types.AutoCCType.NONE,
    },
    {
        'name': 'all',
        'value': data_types.AutoCCType.ALL,
    },
    {
        'name': 'security',
        'value': data_types.AutoCCType.SECURITY,
    },
]


def get_value_by_name(item_list, name):
  """Return value for entry whose name matches the one in item list."""
  for item in item_list:
    if item['name'] == name:
      return item['value']

  return None


class Handler(base_handler.Handler):
  """Configuration manager."""

  @handler.get(handler.HTML)
  @handler.check_admin_access
  def get(self):
    """Handle a get request."""
    external_user_permissions = list(
        data_types.ExternalUserPermission.query().order(
            data_types.ExternalUserPermission.entity_kind,
            data_types.ExternalUserPermission.entity_name,
            data_types.ExternalUserPermission.email))

    template_values = {
        'config': db_config.get(),
        'permissions': external_user_permissions,
        'fieldValues': {
            'csrf_token': form.generate_csrf_token(),
            'user_permission_entity_kinds': USER_PERMISSION_ENTITY_KINDS,
            'user_permission_auto_cc_types': USER_PERMISSION_AUTO_CC_TYPES,
            'add_permission_url': '/add-external-user-permission',
            'delete_permission_url': '/delete-external-user-permission',
        }
    }

    helpers.log('Configuration', helpers.VIEW_OPERATION)
    return self.render('configuration.html', template_values)

  @handler.post(handler.FORM, handler.HTML)
  @handler.check_admin_access
  @handler.require_csrf_token
  def post(self):
    """Handle a post request."""
    config = db_config.get()
    if not config:
      config = data_types.Config()

    previous_hash = request.get('previous_hash')
    if config.previous_hash and config.previous_hash != previous_hash:
      raise helpers.EarlyExitException(
          'Your change conflicts with another configuration update. '
          'Please refresh and try again.', 500)

    build_apiary_service_account_private_key = request.get(
        'build_apiary_service_account_private_key')
    bug_report_url = request.get('bug_report_url')
    client_credentials = request.get('client_credentials')
    jira_url = request.get('jira_url')
    jira_credentials = request.get('jira_credentials')
    component_repository_mappings = request.get('component_repository_mappings')
    contact_string = request.get('contact_string')
    documentation_url = request.get('documentation_url')
    github_credentials = request.get('github_credentials')
    platform_group_mappings = request.get('platform_group_mappings')
    privileged_users = request.get('privileged_users')
    blacklisted_users = request.get('blacklisted_users')
    relax_security_bug_restrictions = request.get(
        'relax_security_bug_restrictions')
    relax_testcase_restrictions = request.get('relax_testcase_restrictions')
    reproduce_tool_client_id = request.get('reproduce_tool_client_id')
    reproduce_tool_client_secret = request.get('reproduce_tool_client_secret')
    reproduction_help_url = request.get('reproduction_help_url')
    test_account_email = request.get('test_account_email')
    test_account_password = request.get('test_account_password')
    wifi_ssid = request.get('wifi_ssid')
    wifi_password = request.get('wifi_password')
    sendgrid_api_key = request.get('sendgrid_api_key')
    sendgrid_sender = request.get('sendgrid_sender')

    config.build_apiary_service_account_private_key = (
        build_apiary_service_account_private_key)
    config.bug_report_url = bug_report_url
    config.client_credentials = client_credentials
    config.component_repository_mappings = component_repository_mappings
    config.contact_string = contact_string
    config.documentation_url = documentation_url
    config.github_credentials = github_credentials
    config.jira_credentials = jira_credentials
    config.jira_url = jira_url
    config.platform_group_mappings = platform_group_mappings
    config.privileged_users = privileged_users
    config.blacklisted_users = blacklisted_users
    config.relax_security_bug_restrictions = bool(
        relax_security_bug_restrictions)
    config.relax_testcase_restrictions = bool(relax_testcase_restrictions)
    config.reproduce_tool_client_id = reproduce_tool_client_id
    config.reproduce_tool_client_secret = reproduce_tool_client_secret
    config.reproduction_help_url = reproduction_help_url
    config.test_account_email = test_account_email
    config.test_account_password = test_account_password
    config.wifi_ssid = wifi_ssid
    config.wifi_password = wifi_password
    config.sendgrid_api_key = sendgrid_api_key
    config.sendgrid_sender = sendgrid_sender

    helpers.log('Configuration', helpers.MODIFY_OPERATION)

    # Before hashing the entity, we must put it so that the internal maps are
    # updated.
    config.put()
    config.previous_hash = utils.entity_hash(config)

    config.put()

    template_values = {
        'title':
            'Success',
        'message': ('Configuration is successfully updated. '
                    'Redirecting to the configuration page...'),
        'redirect_url':
            '/configuration',
    }
    return self.render('message.html', template_values)


class AddExternalUserPermission(base_handler.Handler):
  """Handles adding a new ExternalUserPermission."""

  @handler.post(handler.FORM, handler.HTML)
  @handler.check_admin_access
  @handler.require_csrf_token
  def post(self):
    """Handle a post request."""
    email = utils.normalize_email(request.get('email'))
    entity_kind = request.get('entity_kind')
    entity_name = request.get('entity_name')
    is_prefix = request.get('is_prefix')
    auto_cc = request.get('auto_cc')

    if not email:
      raise helpers.EarlyExitException('No email provided.', 400)

    if not entity_kind or entity_kind == 'undefined':
      raise helpers.EarlyExitException('No entity_kind provided.', 400)

    entity_kind = get_value_by_name(USER_PERMISSION_ENTITY_KINDS, entity_kind)
    if entity_kind is None:
      raise helpers.EarlyExitException('Invalid entity_kind provided.', 400)

    if entity_kind == data_types.PermissionEntityKind.UPLOADER:
      # Enforce null values for entity name and auto-cc when uploader is chosen.
      entity_name = None
      auto_cc = data_types.AutoCCType.NONE
    else:
      if not entity_name:
        raise helpers.EarlyExitException('No entity_name provided.', 400)

      if not auto_cc or auto_cc == 'undefined':
        raise helpers.EarlyExitException('No auto_cc provided.', 400)

      auto_cc = get_value_by_name(USER_PERMISSION_AUTO_CC_TYPES, auto_cc)
      if auto_cc is None:
        raise helpers.EarlyExitException('Invalid auto_cc provided.', 400)

    # Check for existing permission.
    query = data_types.ExternalUserPermission.query(
        data_types.ExternalUserPermission.email == email,
        data_types.ExternalUserPermission.entity_kind == entity_kind,
        data_types.ExternalUserPermission.entity_name == entity_name)

    permission = query.get()
    if not permission:
      # Doesn't exist, create new one.
      permission = data_types.ExternalUserPermission(
          email=email, entity_kind=entity_kind, entity_name=entity_name)

    permission.is_prefix = bool(is_prefix)
    permission.auto_cc = auto_cc
    permission.put()

    helpers.log('Configuration', helpers.MODIFY_OPERATION)
    template_values = {
        'title':
            'Success',
        'message':
            ('User %s permission for entity %s is successfully added. '
             'Redirecting to the configuration page...') % (email, entity_name),
        'redirect_url':
            '/configuration',
    }
    return self.render('message.html', template_values)


class DeleteExternalUserPermission(base_handler.Handler):
  """Handles deleting an ExternalUserPermission."""

  @handler.post(handler.FORM, handler.HTML)
  @handler.check_admin_access
  @handler.require_csrf_token
  def post(self):
    """Handle a post request."""
    email = request.get('email')
    entity_kind = request.get('entity_kind')
    entity_name = request.get('entity_name')

    if not email:
      raise helpers.EarlyExitException('No email provided.', 400)

    if not entity_kind or entity_kind == 'undefined':
      raise helpers.EarlyExitException('No entity_kind provided.', 400)

    entity_kind = get_value_by_name(USER_PERMISSION_ENTITY_KINDS, entity_kind)
    if entity_kind is None:
      raise helpers.EarlyExitException('Invalid entity_kind provided.', 400)

    if entity_kind == data_types.PermissionEntityKind.UPLOADER:
      entity_name = None
    else:
      if not entity_name:
        raise helpers.EarlyExitException('No entity_name provided.', 400)

    # Check for existing permission.
    permission = data_types.ExternalUserPermission.query(
        data_types.ExternalUserPermission.email == email,
        data_types.ExternalUserPermission.entity_kind == entity_kind,
        data_types.ExternalUserPermission.entity_name == entity_name).get()
    if not permission:
      raise helpers.EarlyExitException('Permission does not exist.', 400)
    permission.key.delete()

    helpers.log('Configuration', helpers.MODIFY_OPERATION)
    template_values = {
        'title':
            'Success',
        'message':
            ('User %s permission for entity %s is successfully deleted. '
             'Redirecting to the configuration page...') % (email, entity_name),
        'redirect_url':
            '/configuration',
    }
    return self.render('message.html', template_values)



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Fuzzer utils."""

import os
import re
import stat
import tempfile

from clusterfuzz._internal.base import utils
from clusterfuzz._internal.metrics import logs
from clusterfuzz._internal.system import environment
from clusterfuzz._internal.system import shell

ALLOWED_FUZZ_TARGET_EXTENSIONS = ['', '.exe', '.par']
FUZZ_TARGET_SEARCH_BYTES = b'LLVMFuzzerTestOneInput'
VALID_TARGET_NAME_REGEX = re.compile(r'^[a-zA-Z0-9_-]+$')
BLOCKLISTED_TARGET_NAME_REGEX = re.compile(r'^(jazzer_driver.*)$')


def is_fuzz_target_local(file_path, file_handle=None):
  """Returns whether |file_path| is a fuzz target binary (local path)."""
  # TODO(hzawawy): Handle syzkaller case.
  filename, file_extension = os.path.splitext(os.path.basename(file_path))
  if not VALID_TARGET_NAME_REGEX.match(filename):
    # Check fuzz target has a valid name (without any special chars).
    return False

  if BLOCKLISTED_TARGET_NAME_REGEX.match(filename):
    # Check fuzz target an explicitly disallowed name (e.g. binaries used for
    # jazzer-based targets).
    return False

  if file_extension not in ALLOWED_FUZZ_TARGET_EXTENSIONS:
    # Ignore files with disallowed extensions (to prevent opening e.g. .zips).
    return False

  if not file_handle and not os.path.exists(file_path):
    # Ignore non-existent files for cases when we don't have a file handle.
    return False

  if filename.endswith('_fuzzer'):
    return True

  # TODO(aarya): Remove this optimization if it does not show up significant
  # savings in profiling results.
  fuzz_target_name_regex = environment.get_value('FUZZER_NAME_REGEX')
  if fuzz_target_name_regex:
    return bool(re.match(fuzz_target_name_regex, filename))

  if os.path.exists(file_path) and not stat.S_ISREG(os.stat(file_path).st_mode):
    # Don't read special files (eg: /dev/urandom).
    logs.log_warn('Tried to read from non-regular file: %s.' % file_path)
    return False

  # Use already provided file handle or open the file.
  local_file_handle = file_handle or open(file_path, 'rb')

  # TODO(metzman): Bound this call so we don't read forever if something went
  # wrong.
  result = utils.search_bytes_in_file(FUZZ_TARGET_SEARCH_BYTES,
                                      local_file_handle)

  if not file_handle:
    # If this local file handle is owned by our function, close it now.
    # Otherwise, it is caller's responsibility.
    local_file_handle.close()

  return result


def get_fuzz_targets_local(path):
  """Get list of fuzz targets paths (local)."""
  fuzz_target_paths = []

  for root, _, files in shell.walk(path):
    for filename in files:
      file_path = os.path.join(root, filename)
      if is_fuzz_target_local(file_path):
        fuzz_target_paths.append(file_path)

  return fuzz_target_paths


def get_fuzz_targets(path):
  """Get list of fuzz targets paths."""
  if environment.is_trusted_host():
    from clusterfuzz._internal.bot.untrusted_runner import file_host
    return file_host.get_fuzz_targets(path)
  return get_fuzz_targets_local(path)


def extract_argument(arguments, prefix, remove=True):
  """Extract argument from arguments."""
  for argument in arguments[:]:
    if argument.startswith(prefix):
      if remove:
        arguments.remove(argument)
      return argument[len(prefix):]

  return None


def get_build_revision():
  """Get build revision."""
  try:
    build_revision = int(environment.get_value('APP_REVISION'))
  except (ValueError, TypeError):
    build_revision = -1

  return build_revision


def get_supporting_file(fuzz_target_path, extension_or_suffix):
  """Get supporting file for a fuzz target with the provided extension."""
  base_fuzz_target_path = fuzz_target_path

  # Strip any known extensions.
  for ext in ALLOWED_FUZZ_TARGET_EXTENSIONS:
    if not ext:
      continue

    if base_fuzz_target_path.endswith(ext):
      base_fuzz_target_path = base_fuzz_target_path[:-len(ext)]
      break

  return base_fuzz_target_path + extension_or_suffix


def get_temp_dir():
  """Return the temp dir."""
  temp_dirname = 'temp-' + str(os.getpid())
  temp_directory = os.path.join(
      environment.get_value('FUZZ_INPUTS_DISK', tempfile.gettempdir()),
      temp_dirname)
  shell.create_directory(temp_directory)
  return temp_directory


def get_file_from_untrusted_worker(worker_file_path):
  """Gets file from an untrusted worker to local. Local file stays in the temp
  folder until the end of task or can be explicitly deleted by the caller."""
  from clusterfuzz._internal.bot.untrusted_runner import file_host

  with tempfile.NamedTemporaryFile(delete=False, dir=get_temp_dir()) as f:
    local_file_path = f.name

  file_host.copy_file_from_worker(worker_file_path, local_file_path)
  return local_file_path


def cleanup():
  """Clean up temporary metadata."""
  shell.remove_directory(get_temp_dir())



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Minimize task for handling testcase minimization."""

import binascii
import functools
import os
import threading
import time
import zipfile

import six

from clusterfuzz._internal.base import errors
from clusterfuzz._internal.base import tasks
from clusterfuzz._internal.base import utils
from clusterfuzz._internal.bot import testcase_manager
from clusterfuzz._internal.bot.fuzzers import engine_common
from clusterfuzz._internal.bot.fuzzers.libFuzzer.engine import LibFuzzerEngine
from clusterfuzz._internal.bot.minimizer import basic_minimizers
from clusterfuzz._internal.bot.minimizer import delta_minimizer
from clusterfuzz._internal.bot.minimizer import errors as minimizer_errors
from clusterfuzz._internal.bot.minimizer import html_minimizer
from clusterfuzz._internal.bot.minimizer import js_minimizer
from clusterfuzz._internal.bot.minimizer import minimizer
from clusterfuzz._internal.bot.tasks import setup
from clusterfuzz._internal.bot.tasks import task_creation
from clusterfuzz._internal.bot.tokenizer.antlr_tokenizer import AntlrTokenizer
from clusterfuzz._internal.bot.tokenizer.grammars.JavaScriptLexer import \
    JavaScriptLexer
from clusterfuzz._internal.build_management import build_manager
from clusterfuzz._internal.crash_analysis import severity_analyzer
from clusterfuzz._internal.crash_analysis.crash_comparer import CrashComparer
from clusterfuzz._internal.crash_analysis.crash_result import CrashResult
from clusterfuzz._internal.datastore import data_handler
from clusterfuzz._internal.datastore import data_types
from clusterfuzz._internal.google_cloud_utils import blobs
from clusterfuzz._internal.metrics import logs
from clusterfuzz._internal.platforms import android
from clusterfuzz._internal.system import environment
from clusterfuzz._internal.system import process_handler
from clusterfuzz._internal.system import shell
from clusterfuzz.fuzz import engine

IPCDUMP_TIMEOUT = 60
COMBINED_IPCDUMP_TIMEOUT = 60 * 3
MAX_DEADLINE_EXCEEDED_ATTEMPTS = 3
MAX_TEMPORARY_FILE_BASENAME_LENGTH = 32
MINIMIZE_SANITIZER_OPTIONS_RETRIES = 3
TOKENS_PER_IPCDUMP = 2000

IPC_MESSAGE_UTIL_EXECUTABLE_FOR_PLATFORM = {
    'LINUX': 'ipc_message_util',
    'WINDOWS': 'ipc_message_util.exe',
}

# These options should not ever be removed during minimization. They might seem
# unneeded when reproducing a given crash, but after the bug is fixed, the lack
# of these options might prevent ClusterFuzz from verifying the fix and closing
# the bug. See https://github.com/google/oss-fuzz/issues/3227 for example.
MANDATORY_OSS_FUZZ_OPTIONS = [
    'silence_unsigned_overflow',
]


class MinimizationPhase(object):
  """Effectively an enum to represent the current phase of minimization."""
  GESTURES = 0
  MAIN_FILE = 1
  FILE_LIST = 2
  RESOURCES = 3
  ARGUMENTS = 4


class TestRunner(object):
  """Helper class for running the same test multiple times."""

  def __init__(self, testcase, file_path, files, input_directory, arguments,
               required_arguments, threads, deadline):
    self.testcase = testcase
    self.file_path = file_path
    self.files = files
    self.input_directory = input_directory
    self.gestures = testcase.gestures
    self.arguments = arguments
    self.threads = threads
    self.deadline = deadline

    self.cleanup_interval = environment.get_value(
        'TESTCASES_BEFORE_STALE_PROCESS_CLEANUP', 1)
    self.timeout = environment.get_value('TEST_TIMEOUT', 10)
    self.full_timeout = self.timeout
    self.last_failing_result = None
    self.required_arguments = set(required_arguments.split())

    self.expected_security_flag = False
    self.is_flaky = False
    self.expected_state = None

    self._profile_lock = threading.Lock()
    self._available_profiles = [True] * threads

    self._result_lock = threading.Lock()
    self._results = []

    self._previous_arguments = None

  def _get_profile_index(self):
    """Get the first available profile directory index."""
    with self._profile_lock:
      for index, is_available in enumerate(self._available_profiles):
        if is_available:
          self._available_profiles[index] = False
          return index

    # Raise an exception rather than running in a bad state.
    raise errors.BadStateError('No profile directories available.')

  def _release_profile(self, index):
    """Mark the specified profile as available."""
    with self._profile_lock:
      self._available_profiles[index] = True

  def _handle_test_result(self, result):
    """Handle a test result, return True on pass (no crash), False on fail."""
    if not result.is_crash():
      return True

    # If we have no crash state, we should not consider this a crash.
    state = result.get_state(symbolized=False)
    if not state:
      return True

    # Even though this was a crash, we want to ignore it if the stack does not
    # have the expected security flag (e.g. expected UAF but got NULL deref).
    if result.is_security_issue() != self.expected_security_flag:
      return True

    # Ignore failures that do not appear to be caused by this issue.
    if not self.is_flaky and state != self.expected_state:
      return True

    self.last_failing_result = result
    return False

  def _repopulate_required_arguments(self, arguments):
    """Add required arguments back to the argument list."""
    fixed_arguments = []
    original_arguments = self.arguments.split()

    original_argument_index = 0
    argument_index = 0

    while original_argument_index < len(original_arguments):
      original_argument = original_arguments[original_argument_index]
      if (argument_index < len(arguments) and
          original_argument == arguments[argument_index]):
        argument_index += 1
        fixed_arguments.append(original_argument)
      elif (original_argument in self.required_arguments or
            original_argument.split('=')[0] in self.required_arguments or
            '"' in original_argument or "'" in original_argument):
        fixed_arguments.append(original_argument)

      original_argument_index += 1

    return fixed_arguments

  def get_argument_string(self, arguments):
    """Convert a list of argument tokens to a usable value."""
    fixed_arguments = self._repopulate_required_arguments(arguments)
    return ' '.join(fixed_arguments)

  def test_with_defaults(self, _):
    """Run a test with all default values."""
    result = self.run()
    return self._handle_test_result(result)

  def test_with_files(self, files):
    """Run the test with the specified file list."""
    files_to_rename = list(set(self.files) - set(files))
    files_to_skip = []

    # Generate a unique suffix to append to files we want to ignore.
    index = 0
    file_rename_suffix = '___%d' % index
    while any([f.endswith(file_rename_suffix) for f in files_to_rename]):
      index += 1
      file_rename_suffix = '___%d' % index

    # Rename all files in the test case's file list but not the specified one.
    for file_to_rename in files_to_rename:
      absolute_file_to_rename = os.path.join(self.input_directory,
                                             file_to_rename)
      try:
        os.rename(absolute_file_to_rename,
                  '%s%s' % (absolute_file_to_rename, file_rename_suffix))
      except OSError:
        # This can happen if we have already renamed a directory with files
        # under it. In this case, make sure we don't try to change the name
        # back later.
        files_to_skip.append(file_to_rename)

    # Clean up any issues with modifications of resources in subdirectories.
    for file_to_skip in files_to_skip:
      files_to_rename.remove(file_to_skip)
    files_to_rename.reverse()

    result = self.run()

    # Restore previously renamed files to their original locations.
    for file_to_rename in files_to_rename:
      absolute_file_to_rename = os.path.join(self.input_directory,
                                             file_to_rename)
      os.rename('%s%s' % (absolute_file_to_rename, file_rename_suffix),
                absolute_file_to_rename)

    return self._handle_test_result(result)

  def test_with_file(self, file_path):
    """Run the test with the specified contents for a particular file."""
    result = self.run(file_path=file_path)
    return self._handle_test_result(result)

  def test_with_gestures(self, gestures):
    """Run the test with the specified gesture list."""
    result = self.run(gestures=gestures)
    return self._handle_test_result(result)

  def test_with_command_line_arguments(self, arguments):
    """Run the test with the specified command line."""
    fixed_arguments = self.get_argument_string(arguments)
    result = self.run(
        arguments=fixed_arguments,
        timeout=self.full_timeout,
        use_fresh_profile=True)
    return self._handle_test_result(result)

  def set_test_expectations(self, security_flag, is_flaky,
                            unsymbolized_crash_state):
    """Set expectations when using this runner for tests."""
    self.expected_security_flag = security_flag
    self.is_flaky = is_flaky
    self.expected_state = unsymbolized_crash_state

  def run(self,
          file_path=None,
          gestures=None,
          arguments=None,
          timeout=None,
          log_command=False,
          use_fresh_profile=False):
    """Run the test."""
    if file_path is None:
      file_path = self.file_path

    if gestures is None:
      gestures = self.gestures

    if arguments is None:
      arguments = self.arguments

    # TODO(mbarbella): Dynamic timeout adjustment.
    if timeout is None:
      timeout = self.timeout

    needs_http = self.testcase.http_flag
    profile_index = self._get_profile_index()

    if use_fresh_profile and environment.get_value('USER_PROFILE_ARG'):
      shell.remove_directory(
          testcase_manager.get_user_profile_directory(profile_index))

    # For Android, we need to sync our local testcases directory with the one on
    # the device.
    if environment.is_android():
      android.device.push_testcases_to_device()
    elif environment.is_trusted_host():
      from clusterfuzz._internal.bot.untrusted_runner import file_host
      file_host.push_testcases_to_worker()

    # If we need to write a command line file, only do so if the arguments have
    # changed.
    arguments_changed = arguments != self._previous_arguments
    self._previous_arguments = arguments

    command = testcase_manager.get_command_line_for_application(
        file_to_run=file_path,
        app_args=arguments,
        needs_http=needs_http,
        user_profile_index=profile_index,
        write_command_line_file=arguments_changed)
    if log_command:
      logs.log('Executing command: %s' % command)

    return_code, crash_time, output = process_handler.run_process(
        command, timeout=timeout, gestures=gestures)

    self._release_profile(profile_index)
    return CrashResult(return_code, crash_time, output)

  def store_result_from_run(self, result):
    """Run and store the result for later processing."""
    with self._result_lock:
      self._results.append(result)

    # A race here isn't problematic. Better not to hold the lock during an
    # is_crash call.
    if not self.last_failing_result and result.is_crash():
      self.last_failing_result = result

  def execute_parallel_runs(self, runs, instances=None):
    """Run multiple instances of this test in parallel."""
    if not instances:
      instances = self.threads

    # TODO(mbarbella): Hack for Android. If we are running single-threaded, it
    # is safe to call a cleanup function on each thread. Ideally, the minimizer
    # would like to assume that when it finishes running a process it cleans
    # itself up properly.
    cleanup_function = None
    if self.threads == 1:
      cleanup_function = process_handler.cleanup_stale_processes

    run_queue = minimizer.TestQueue(
        instances, per_thread_cleanup_function=cleanup_function)
    for _ in range(runs):
      run_queue.push(self.file_path, self.run, self.store_result_from_run)

    run_queue.process()

    # At timeout, we send SIGTERM. Wait for 2 seconds before sending SIGKILL.
    time.sleep(2)
    process_handler.cleanup_stale_processes()

    with self._result_lock:
      results = self._results
      self._results = []

    return results


def execute_task(testcase_id, job_type):
  """Attempt to minimize a given testcase."""
  # Get deadline to finish this task.
  deadline = tasks.get_task_completion_deadline()

  # Locate the testcase associated with the id.
  testcase = data_handler.get_testcase_by_id(testcase_id)
  if not testcase:
    return

  # Update comments to reflect bot information.
  data_handler.update_testcase_comment(testcase, data_types.TaskState.STARTED)

  # Setup testcase and its dependencies. Also, allow setting up a different
  # fuzzer.
  minimize_fuzzer_override = environment.get_value('MINIMIZE_FUZZER_OVERRIDE')
  file_list, input_directory, testcase_file_path = setup.setup_testcase(
      testcase, job_type, fuzzer_override=minimize_fuzzer_override)
  if not file_list:
    return

  # Initialize variables.
  max_timeout = environment.get_value('TEST_TIMEOUT', 10)
  app_arguments = environment.get_value('APP_ARGS')

  # Set up a custom or regular build based on revision.
  last_tested_crash_revision = testcase.get_metadata(
      'last_tested_crash_revision')

  crash_revision = last_tested_crash_revision or testcase.crash_revision
  build_manager.setup_build(crash_revision)

  # Check if we have an application path. If not, our build failed
  # to setup correctly.
  if not build_manager.check_app_path():
    logs.log_error('Unable to setup build for minimization.')
    build_fail_wait = environment.get_value('FAIL_WAIT')

    if environment.get_value('ORIGINAL_JOB_NAME'):
      _skip_minimization(testcase, 'Failed to setup build for overridden job.')
    else:
      # Only recreate task if this isn't an overriden job. It's possible that a
      # revision exists for the original job, but doesn't exist for the
      # overriden job.
      tasks.add_task(
          'minimize', testcase_id, job_type, wait_time=build_fail_wait)

    return

  if environment.is_libfuzzer_job():
    do_libfuzzer_minimization(testcase, testcase_file_path)
    return

  if environment.is_engine_fuzzer_job():
    # TODO(ochang): More robust check for engine minimization support.
    _skip_minimization(testcase, 'Engine does not support minimization.')
    return

  max_threads = utils.maximum_parallel_processes_allowed()

  # Prepare the test case runner.
  crash_retries = environment.get_value('CRASH_RETRIES')
  warmup_timeout = environment.get_value('WARMUP_TIMEOUT')
  required_arguments = environment.get_value('REQUIRED_APP_ARGS', '')

  # Add any testcase-specific required arguments if needed.
  additional_required_arguments = testcase.get_metadata(
      'additional_required_app_args')
  if additional_required_arguments:
    required_arguments = '%s %s' % (required_arguments,
                                    additional_required_arguments)

  test_runner = TestRunner(testcase, testcase_file_path, file_list,
                           input_directory, app_arguments, required_arguments,
                           max_threads, deadline)

  # Verify the crash with a long timeout.
  warmup_crash_occurred = False
  result = test_runner.run(timeout=warmup_timeout, log_command=True)
  if result.is_crash():
    warmup_crash_occurred = True
    logs.log('Warmup crash occurred in %d seconds.' % result.crash_time)

  saved_unsymbolized_crash_state, flaky_stack, crash_times = (
      check_for_initial_crash(test_runner, crash_retries, testcase))

  # If the warmup crash occurred but we couldn't reproduce this in with
  # multiple processes running in parallel, try to minimize single threaded.
  reproducible_crash_count = (
      testcase_manager.REPRODUCIBILITY_FACTOR * crash_retries)
  if (len(crash_times) < reproducible_crash_count and warmup_crash_occurred and
      max_threads > 1):
    logs.log('Attempting to continue single-threaded.')

    max_threads = 1
    test_runner = TestRunner(testcase, testcase_file_path, file_list,
                             input_directory, app_arguments, required_arguments,
                             max_threads, deadline)

    saved_unsymbolized_crash_state, flaky_stack, crash_times = (
        check_for_initial_crash(test_runner, crash_retries, testcase))

  if not crash_times:
    # We didn't crash at all. This might be a legitimately unreproducible
    # test case, so it will get marked as such after being retried on other
    # bots.
    testcase = data_handler.get_testcase_by_id(testcase_id)
    data_handler.update_testcase_comment(testcase, data_types.TaskState.ERROR,
                                         'Unable to reproduce crash')
    task_creation.mark_unreproducible_if_flaky(testcase, True)
    return

  if flaky_stack:
    testcase = data_handler.get_testcase_by_id(testcase_id)
    testcase.flaky_stack = flaky_stack
    testcase.put()

  is_redo = testcase.get_metadata('redo_minimize')
  if not is_redo and len(crash_times) < reproducible_crash_count:
    # We reproduced this crash at least once. It's too flaky to minimize, but
    # maybe we'll have more luck in the other jobs.
    testcase = data_handler.get_testcase_by_id(testcase_id)
    testcase.minimized_keys = 'NA'
    error_message = (
        'Crash occurs, but not too consistently. Skipping minimization '
        '(crashed %d/%d)' % (len(crash_times), crash_retries))
    data_handler.update_testcase_comment(testcase, data_types.TaskState.ERROR,
                                         error_message)
    create_additional_tasks(testcase)
    return

  # If we've made it this far, the test case appears to be reproducible. Clear
  # metadata from previous runs had it been marked as potentially flaky.
  task_creation.mark_unreproducible_if_flaky(testcase, False)

  test_runner.set_test_expectations(testcase.security_flag, flaky_stack,
                                    saved_unsymbolized_crash_state)

  # Use the max crash time unless this would be greater than the max timeout.
  test_timeout = min(max(crash_times), max_timeout) + 1
  logs.log('Using timeout %d (was %d)' % (test_timeout, max_timeout))
  test_runner.timeout = test_timeout

  logs.log('Starting minimization.')

  if should_attempt_phase(testcase, MinimizationPhase.GESTURES):
    gestures = minimize_gestures(test_runner, testcase)

    # We can't call check_deadline_exceeded_and_store_partial_minimized_testcase
    # at this point because we do not have a test case to store.
    testcase = data_handler.get_testcase_by_id(testcase.key.id())

    if testcase.security_flag and len(testcase.gestures) != len(gestures):
      # Re-run security severity analysis since gestures affect the severity.
      testcase.security_severity = severity_analyzer.get_security_severity(
          testcase.crash_type, data_handler.get_stacktrace(testcase), job_type,
          bool(gestures))

    testcase.gestures = gestures
    testcase.set_metadata('minimization_phase', MinimizationPhase.MAIN_FILE)

    if time.time() > test_runner.deadline:
      tasks.add_task('minimize', testcase.key.id(), job_type)
      return

  # Minimize the main file.
  data = utils.get_file_contents_with_fatal_error_on_failure(testcase_file_path)
  if should_attempt_phase(testcase, MinimizationPhase.MAIN_FILE):
    data = minimize_main_file(test_runner, testcase_file_path, data)

    if check_deadline_exceeded_and_store_partial_minimized_testcase(
        deadline, testcase_id, job_type, input_directory, file_list, data,
        testcase_file_path):
      return

    testcase.set_metadata('minimization_phase', MinimizationPhase.FILE_LIST)

  # Minimize the file list.
  if should_attempt_phase(testcase, MinimizationPhase.FILE_LIST):
    if environment.get_value('MINIMIZE_FILE_LIST', True):
      file_list = minimize_file_list(test_runner, file_list, input_directory,
                                     testcase_file_path)

      if check_deadline_exceeded_and_store_partial_minimized_testcase(
          deadline, testcase_id, job_type, input_directory, file_list, data,
          testcase_file_path):
        return
    else:
      logs.log('Skipping minimization of file list.')

    testcase.set_metadata('minimization_phase', MinimizationPhase.RESOURCES)

  # Minimize any files remaining in the file list.
  if should_attempt_phase(testcase, MinimizationPhase.RESOURCES):
    if environment.get_value('MINIMIZE_RESOURCES', True):
      for dependency in file_list:
        minimize_resource(test_runner, dependency, input_directory,
                          testcase_file_path)

        if check_deadline_exceeded_and_store_partial_minimized_testcase(
            deadline, testcase_id, job_type, input_directory, file_list, data,
            testcase_file_path):
          return
    else:
      logs.log('Skipping minimization of resources.')

    testcase.set_metadata('minimization_phase', MinimizationPhase.ARGUMENTS)

  if should_attempt_phase(testcase, MinimizationPhase.ARGUMENTS):
    app_arguments = minimize_arguments(test_runner, app_arguments)

    # Arguments must be stored here in case we time out below.
    testcase.minimized_arguments = app_arguments
    testcase.put()

    if check_deadline_exceeded_and_store_partial_minimized_testcase(
        deadline, testcase_id, job_type, input_directory, file_list, data,
        testcase_file_path):
      return

  command = testcase_manager.get_command_line_for_application(
      testcase_file_path, app_args=app_arguments, needs_http=testcase.http_flag)
  last_crash_result = test_runner.last_failing_result

  store_minimized_testcase(testcase, input_directory, file_list, data,
                           testcase_file_path)
  finalize_testcase(
      testcase_id, command, last_crash_result, flaky_stack=flaky_stack)


def finalize_testcase(testcase_id,
                      command,
                      last_crash_result,
                      flaky_stack=False):
  """Perform final updates on a test case and prepare it for other tasks."""
  # Symbolize crash output if we have it.
  testcase = data_handler.get_testcase_by_id(testcase_id)
  if last_crash_result:
    _update_crash_result(testcase, last_crash_result, command)
  testcase.delete_metadata('redo_minimize', update_testcase=False)

  # Update remaining test case information.
  testcase.flaky_stack = flaky_stack
  if build_manager.is_custom_binary():
    testcase.set_impacts_as_na()
    testcase.regression = 'NA'
  data_handler.update_testcase_comment(testcase, data_types.TaskState.FINISHED)

  # We might have updated the crash state. See if we need to marked as duplicate
  # based on other testcases.
  data_handler.handle_duplicate_entry(testcase)

  create_additional_tasks(testcase)


def create_additional_tasks(testcase):
  """Create post-minimization tasks for this reproducible testcase such as
  impact, regression, progression, variant and symbolize."""
  # No need to create progression task. It is automatically created by the cron
  # handler.
  task_creation.create_impact_task_if_needed(testcase)
  task_creation.create_regression_task_if_needed(testcase)
  task_creation.create_symbolize_task_if_needed(testcase)
  task_creation.create_variant_tasks_if_needed(testcase)


def should_attempt_phase(testcase, phase):
  """Return true if we should we attempt a minimization phase."""
  if (phase == MinimizationPhase.ARGUMENTS and
      environment.is_engine_fuzzer_job()):
    # Should not minimize arguments list for engine based fuzzer jobs.
    return False

  current_phase = testcase.get_metadata(
      'minimization_phase', default=MinimizationPhase.GESTURES)
  return phase >= current_phase


def minimize_gestures(test_runner, testcase):
  """Minimize the gesture list for a test case."""
  gestures = testcase.gestures
  if gestures:
    gesture_minimizer = delta_minimizer.DeltaMinimizer(
        test_runner.test_with_gestures,
        max_threads=test_runner.threads,
        tokenize=False,
        deadline=test_runner.deadline,
        cleanup_function=process_handler.cleanup_stale_processes,
        single_thread_cleanup_interval=test_runner.cleanup_interval,
        progress_report_function=functools.partial(logs.log))
    gestures = gesture_minimizer.minimize(gestures)

  logs.log('Minimized gestures: %s' % str(gestures))
  return gestures


def minimize_main_file(test_runner, testcase_file_path, data):
  """Minimize the main test case file."""
  if not can_minimize_file(testcase_file_path):
    return data

  get_random_file = functools.partial(get_temporary_file, testcase_file_path)
  data = (
      minimize_file(testcase_file_path, test_runner.test_with_file,
                    get_random_file, data, test_runner.deadline,
                    test_runner.threads, test_runner.cleanup_interval))

  logs.log('Minimized main test file.')
  return data


def minimize_file_list(test_runner, file_list, input_directory, main_file):
  """Minimize the test case files."""
  if len(file_list) <= 1:
    return file_list

  # TODO(mbarbella): Simplify this with refactoring of setup_testcase.
  offset = len(input_directory) + len(os.path.sep)
  fixed_testcase_file_path = main_file[offset:]

  # As of now, this must be single-threaded.
  file_list_minimizer = basic_minimizers.SinglePassMinimizer(
      test_runner.test_with_files,
      tokenize=False,
      deadline=test_runner.deadline,
      cleanup_function=process_handler.cleanup_stale_processes,
      single_thread_cleanup_interval=test_runner.cleanup_interval,
      progress_report_function=functools.partial(logs.log))
  file_list = file_list_minimizer.minimize(file_list)

  if fixed_testcase_file_path not in file_list:
    file_list.append(fixed_testcase_file_path)

  logs.log('Minimized file list: %s' % str(file_list))
  return file_list


def minimize_resource(test_runner, dependency, input_directory, main_file):
  """Minimize a resource for the test case."""
  # TODO(mbarbella): Simplify this with refactoring of setup_testcase.
  offset = len(input_directory) + len(os.path.sep)
  fixed_testcase_file_path = main_file[offset:]

  dependency_absolute_path = os.path.join(input_directory, dependency)

  if (dependency == fixed_testcase_file_path or dependency == main_file or
      not can_minimize_file(dependency_absolute_path)):
    return

  get_temp_file = functools.partial(
      get_temporary_file, dependency_absolute_path, no_modifications=True)
  original_data = utils.get_file_contents_with_fatal_error_on_failure(
      dependency_absolute_path)
  dependency_data = (
      minimize_file(
          dependency,
          test_runner.test_with_defaults,
          get_temp_file,
          original_data,
          test_runner.deadline,
          1,
          test_runner.cleanup_interval,
          delete_temp_files=False))
  utils.write_data_to_file(dependency_data, dependency_absolute_path)

  logs.log('Minimized dependency file: %s' % dependency)


def minimize_arguments(test_runner, app_arguments):
  """Minimize the argument list for a test case."""
  argument_minimizer = delta_minimizer.DeltaMinimizer(
      test_runner.test_with_command_line_arguments,
      max_threads=test_runner.threads,
      tokenize=False,
      deadline=test_runner.deadline,
      cleanup_function=process_handler.cleanup_stale_processes,
      single_thread_cleanup_interval=test_runner.cleanup_interval,
      progress_report_function=functools.partial(logs.log))
  reduced_args = argument_minimizer.minimize(app_arguments.split())
  reduced_arg_string = test_runner.get_argument_string(reduced_args)

  return reduced_arg_string


def store_minimized_testcase(testcase, base_directory, file_list,
                             file_to_run_data, file_to_run):
  """Store all files that make up this testcase."""
  # Write the main file data.
  utils.write_data_to_file(file_to_run_data, file_to_run)

  # Prepare the file.
  zip_path = None
  if testcase.archive_state:
    if len(file_list) > 1:
      testcase.archive_state |= data_types.ArchiveStatus.MINIMIZED
      zip_path = os.path.join(
          environment.get_value('INPUT_DIR'), '%d.zip' % testcase.key.id())
      zip_file = zipfile.ZipFile(zip_path, 'w')
      count = 0
      filtered_file_list = []
      for file_name in file_list:
        absolute_filename = os.path.join(base_directory, file_name)
        is_file = os.path.isfile(absolute_filename)
        if file_to_run_data and is_file and os.path.getsize(
            absolute_filename) == 0 and (os.path.basename(
                absolute_filename).encode('utf-8') not in file_to_run_data):
          continue
        if not os.path.exists(absolute_filename):
          continue
        zip_file.write(absolute_filename, file_name, zipfile.ZIP_DEFLATED)
        if is_file:
          count += 1
          filtered_file_list.append(absolute_filename)

      zip_file.close()
      try:
        if count > 1:
          file_handle = open(zip_path, 'rb')
        else:
          if not filtered_file_list:
            # We minimized everything. The only thing needed to reproduce is the
            # interaction gesture.
            file_path = file_list[0]
            file_handle = open(file_path, 'wb')
            file_handle.close()
          else:
            file_path = filtered_file_list[0]
          file_handle = open(file_path, 'rb')
          testcase.absolute_path = os.path.join(base_directory,
                                                os.path.basename(file_path))
          testcase.archive_state &= ~data_types.ArchiveStatus.MINIMIZED
      except IOError:
        testcase.put()  # Preserve what we can.
        logs.log_error('Unable to open archive for blobstore write.')
        return
    else:
      absolute_filename = os.path.join(base_directory, file_list[0])
      file_handle = open(absolute_filename, 'rb')
      testcase.archive_state &= ~data_types.ArchiveStatus.MINIMIZED
  else:
    file_handle = open(file_list[0], 'rb')
    testcase.archive_state &= ~data_types.ArchiveStatus.MINIMIZED

  # Store the testcase.
  minimized_keys = blobs.write_blob(file_handle)
  file_handle.close()

  testcase.minimized_keys = minimized_keys
  testcase.put()

  if zip_path:
    shell.remove_file(zip_path)


def check_deadline_exceeded_and_store_partial_minimized_testcase(
    deadline, testcase_id, job_type, input_directory, file_list,
    file_to_run_data, main_file_path):
  """Store the partially minimized test and check the deadline."""
  testcase = data_handler.get_testcase_by_id(testcase_id)
  store_minimized_testcase(testcase, input_directory, file_list,
                           file_to_run_data, main_file_path)

  deadline_exceeded = time.time() > deadline
  if deadline_exceeded:
    attempts = testcase.get_metadata(
        'minimization_deadline_exceeded_attempts', default=0)
    if attempts >= MAX_DEADLINE_EXCEEDED_ATTEMPTS:
      _skip_minimization(testcase,
                         'Exceeded minimization deadline too many times.')
    else:
      testcase.set_metadata('minimization_deadline_exceeded_attempts',
                            attempts + 1)
      tasks.add_task('minimize', testcase_id, job_type)

  return deadline_exceeded


def check_for_initial_crash(test_runner, crash_retries, testcase):
  """Initial check to see how long it takes to reproduce a crash."""
  crash_times = []
  flaky_stack = False
  saved_crash_state = None
  saved_security_flag = None
  saved_unsymbolized_crash_state = None

  results = test_runner.execute_parallel_runs(crash_retries)

  for result in results:
    if not result.is_crash():
      continue

    if result.should_ignore():
      continue

    crash_state = result.get_state(symbolized=True)
    security_flag = result.is_security_issue()
    unsymbolized_crash_state = result.get_state(symbolized=False)

    if not unsymbolized_crash_state:
      continue

    if security_flag != testcase.security_flag:
      continue

    crash_times.append(result.crash_time)

    if not saved_crash_state:
      saved_crash_state = crash_state
      saved_security_flag = security_flag
      saved_unsymbolized_crash_state = unsymbolized_crash_state
      continue

    crash_comparer = CrashComparer(crash_state, saved_crash_state)
    if not crash_comparer.is_similar():
      flaky_stack = True

  logs.log('Total crash count: %d/%d. Flaky: %s. Security: %s. State:\n%s' %
           (len(crash_times), crash_retries, flaky_stack, saved_security_flag,
            saved_crash_state))

  return saved_unsymbolized_crash_state, flaky_stack, crash_times


def get_temporary_file_name(original_file):
  """Generate a temporary file name in the same directory as |original_file|."""
  directory, basename = os.path.split(original_file)
  basename = basename[-MAX_TEMPORARY_FILE_BASENAME_LENGTH:]

  random_hex = binascii.b2a_hex(os.urandom(16)).decode('utf-8')
  new_file_path = os.path.join(directory, '%s%s' % (random_hex, basename))

  return new_file_path


def get_temporary_file(original_file, no_modifications=False):
  """Get a temporary file handle with a name based on an original file name."""
  if no_modifications:
    handle = open(original_file, 'wb')
    return handle

  handle = open(get_temporary_file_name(original_file), 'wb')
  return handle


def get_ipc_message_util_executable():
  """Return the ipc_message_util executable path for the current build."""
  app_directory = environment.get_value('APP_DIR')
  platform = environment.platform()

  try:
    executable = IPC_MESSAGE_UTIL_EXECUTABLE_FOR_PLATFORM[platform]
  except KeyError:
    # Current platform is not supported.
    return None

  return os.path.join(app_directory, executable)


def create_partial_ipc_dump(tokens, original_file_path):
  """Use the ipc_message_util utility to create a file for up to
     |TOKENS_PER_IPCDUMP| tokens."""
  assert len(tokens) <= TOKENS_PER_IPCDUMP

  token_list = ','.join([str(token) for token in tokens])
  temp_file_path = get_temporary_file_name(original_file_path)

  executable = get_ipc_message_util_executable()
  command_line = shell.get_command_line_from_argument_list(
      [executable,
       '--in=%s' % token_list, original_file_path, temp_file_path])
  return_code, _, output = process_handler.run_process(
      command_line, testcase_run=False, timeout=IPCDUMP_TIMEOUT)
  if return_code or not os.path.exists(temp_file_path):
    # For some reason, generating the new file failed.
    logs.log_error('Failed to create ipc dump file %s.' % output)
    return None

  return temp_file_path


def combine_ipc_dumps(ipcdumps, original_file_path):
  """Combines a list of ipcdump files into a single dump."""
  input_file_string = ','.join(ipcdumps)
  executable = get_ipc_message_util_executable()
  output_file_path = get_temporary_file_name(original_file_path)
  command_line = shell.get_command_line_from_argument_list(
      [executable, input_file_string, output_file_path])
  return_code, _, output = process_handler.run_process(
      command_line, testcase_run=False, timeout=COMBINED_IPCDUMP_TIMEOUT)

  for ipcdump in ipcdumps:
    shell.remove_file(ipcdump)

  if return_code or not os.path.exists(output_file_path):
    logs.log_error('Failed to create ipc dump file %s.' % output)
    return None

  return output_file_path


def supports_ipc_minimization(file_path):
  """Check to see if IPC minimization is supported for the current build."""
  executable = get_ipc_message_util_executable()
  if not executable:
    # IPC fuzzer minimization is not supported on this platform.
    return False

  command_line = shell.get_command_line_from_argument_list(
      [executable, '--dump', '--in=0', file_path])
  return_code, _, output = process_handler.run_process(
      command_line, testcase_run=False, timeout=IPCDUMP_TIMEOUT)

  # If --in is not supported by this version of the ipc_message_util binary,
  # it will exit with a nonzero exit status. Also ensure that the first message
  # is printed in case the build is bad for some other reason.
  # Example output: 0. AutofillHostMsg_DidFillAutofillFormData
  if return_code or not output.startswith('0.'):
    return False

  supports_ipc_minimization.is_supported = True
  return True


def can_minimize_file(file_path):
  """Check to see if we support minimization for this file."""
  # If this is not a binary file, we should be able to minimize it in some way.
  if not utils.is_binary_file(file_path):
    return True

  # Attempt to minimize IPC dumps.
  if file_path.endswith(testcase_manager.IPCDUMP_EXTENSION):
    return supports_ipc_minimization(file_path)

  # Other binary file formats are not supported.
  return False


def do_ipc_dump_minimization(test_function, get_temp_file, file_path, deadline,
                             threads, cleanup_interval, delete_temp_files):
  """IPC dump minimization strategy."""

  def tokenize(current_file_path):
    """Generate a token list for an IPC fuzzer test case."""
    command_line = shell.get_command_line_from_argument_list(
        [get_ipc_message_util_executable(), '--dump', current_file_path])
    _, _, output = process_handler.run_process(
        command_line, testcase_run=False, timeout=IPCDUMP_TIMEOUT)
    output_lines = output.splitlines()
    if not output_lines:
      return []

    # Each output line starts with the message index followed by a ".", but
    # we are only interested in the total number of messages in the file. To
    # find this, we add one to the index of the final message.
    try:
      last_index = int(output_lines[-1].split('.')[0])
    except ValueError:
      return []

    return list(range(last_index + 1))

  def combine_tokens(tokens):
    """Use the ipc_message_util utility to create a file for these tokens."""
    partial_ipcdumps = []
    for start_index in range(0, len(tokens), TOKENS_PER_IPCDUMP):
      end_index = min(start_index + TOKENS_PER_IPCDUMP, len(tokens))
      current_tokens = tokens[start_index:end_index]
      partial_ipcdumps.append(
          create_partial_ipc_dump(current_tokens, file_path))

    combined_file_path = None
    if len(partial_ipcdumps) > 1:
      combined_file_path = combine_ipc_dumps(partial_ipcdumps, file_path)
    elif len(partial_ipcdumps) == 1:
      combined_file_path = partial_ipcdumps[0]

    if not combined_file_path:
      # This can happen in the case of a timeout or other error. The actual
      # error should already be logged, so no need to do it again here.
      return b''

    # TODO(mbarbella): Allow token combining functions to write files directly.
    handle = open(combined_file_path, 'rb')
    result = handle.read()
    handle.close()

    shell.remove_file(combined_file_path)
    return result

  current_minimizer = delta_minimizer.DeltaMinimizer(
      test_function,
      max_threads=threads,
      deadline=deadline,
      cleanup_function=process_handler.cleanup_stale_processes,
      single_thread_cleanup_interval=cleanup_interval,
      get_temp_file=get_temp_file,
      delete_temp_files=delete_temp_files,
      tokenizer=tokenize,
      token_combiner=combine_tokens,
      progress_report_function=functools.partial(logs.log))
  return current_minimizer.minimize(file_path)


def do_js_minimization(test_function, get_temp_file, data, deadline, threads,
                       cleanup_interval, delete_temp_files):
  """Javascript minimization strategy."""
  # Start by using a generic line minimizer on the test.
  # Do two line minimizations to make up for the fact that minimzations on bots
  # don't always minimize as much as they can.
  for _ in range(2):
    data = do_line_minimization(test_function, get_temp_file, data, deadline,
                                threads, cleanup_interval, delete_temp_files)

  tokenizer = AntlrTokenizer(JavaScriptLexer)

  current_minimizer = js_minimizer.JSMinimizer(
      test_function,
      max_threads=threads,
      deadline=deadline,
      cleanup_function=process_handler.cleanup_stale_processes,
      single_thread_cleanup_interval=cleanup_interval,
      get_temp_file=get_temp_file,
      delete_temp_files=delete_temp_files,
      tokenizer=tokenizer.tokenize,
      token_combiner=tokenizer.combine,
      progress_report_function=functools.partial(logs.log))

  # Some tokens can't be removed until other have, so do 2 passes.
  try:
    for _ in range(2):
      data = current_minimizer.minimize(data)
  except minimizer_errors.AntlrDecodeError:
    data = do_line_minimization(test_function, get_temp_file, data, deadline,
                                threads, cleanup_interval, delete_temp_files)

  # FIXME(mbarbella): Improve the JS minimizer so that this is not necessary.
  # Sometimes, lines that could not have been removed on their own can now be
  # removed since they have already been partially cleaned up.
  return do_line_minimization(test_function, get_temp_file, data, deadline,
                              threads, cleanup_interval, delete_temp_files)


def _run_libfuzzer_testcase(testcase, testcase_file_path, crash_retries=1):
  """Run libFuzzer testcase, and return the CrashResult."""
  # Cleanup any existing application instances and temp directories.
  process_handler.cleanup_stale_processes()
  shell.clear_temp_directory()

  if environment.is_trusted_host():
    from clusterfuzz._internal.bot.untrusted_runner import file_host
    file_host.copy_file_to_worker(
        testcase_file_path, file_host.rebase_to_worker_root(testcase_file_path))

  test_timeout = environment.get_value('TEST_TIMEOUT',
                                       process_handler.DEFAULT_TEST_TIMEOUT)
  return testcase_manager.test_for_crash_with_retries(
      testcase,
      testcase_file_path,
      test_timeout,
      compare_crash=False,
      crash_retries=crash_retries)


def run_libfuzzer_engine(tool_name, target_name, arguments, testcase_path,
                         output_path, timeout):
  """Run the libFuzzer engine."""
  arguments = list(arguments)
  if environment.is_trusted_host():
    from clusterfuzz._internal.bot.untrusted_runner import tasks_host

    # TODO(ochang): Remove hardcode.
    return tasks_host.process_testcase('libFuzzer', tool_name, target_name,
                                       arguments, testcase_path, output_path,
                                       timeout)

  target_path = engine_common.find_fuzzer_path(
      environment.get_value('BUILD_DIR'), target_name)
  if not target_path:
    return engine.ReproduceResult([], 0, 0, '')

  engine_impl = LibFuzzerEngine()
  if tool_name == 'minimize':
    func = engine_impl.minimize_testcase
  else:
    assert tool_name == 'cleanse'
    func = engine_impl.cleanse

  return func(target_path, arguments, testcase_path, output_path, timeout)


def _run_libfuzzer_tool(tool_name,
                        testcase,
                        testcase_file_path,
                        timeout,
                        expected_crash_state,
                        set_dedup_flags=False):
  """Run libFuzzer tool to either minimize or cleanse."""
  memory_tool_options_var = environment.get_current_memory_tool_var()
  saved_memory_tool_options = environment.get_value(memory_tool_options_var)

  def _set_dedup_flags():
    """Allow libFuzzer to do its own crash comparison during minimization."""
    memory_tool_options = environment.get_memory_tool_options(
        memory_tool_options_var, default_value={})

    memory_tool_options['symbolize'] = 1
    memory_tool_options['dedup_token_length'] = 3

    environment.set_memory_tool_options(memory_tool_options_var,
                                        memory_tool_options)

  def _unset_dedup_flags():
    """Reset memory tool options."""
    # This is needed so that when we re-run, we can symbolize ourselves
    # (ignoring inline frames).
    if saved_memory_tool_options is not None:
      environment.set_value(memory_tool_options_var, saved_memory_tool_options)

  output_file_path = get_temporary_file_name(testcase_file_path)

  arguments = data_handler.get_arguments(testcase).split()
  fuzzer_display = data_handler.get_fuzzer_display(testcase)

  if set_dedup_flags:
    _set_dedup_flags()

  try:
    result = run_libfuzzer_engine(tool_name, fuzzer_display.target, arguments,
                                  testcase_file_path, output_file_path, timeout)
  except TimeoutError:
    logs.log_warn('LibFuzzer timed out.')
    return None, None

  if set_dedup_flags:
    _unset_dedup_flags()

  if not os.path.exists(output_file_path):
    logs.log_warn('LibFuzzer %s run failed.' % tool_name, output=result.output)
    return None, None

  # Ensure that the crash parameters match. It's possible that we will
  # minimize/cleanse to an unrelated bug, such as a timeout.
  crash_result = _run_libfuzzer_testcase(testcase, output_file_path)
  state = crash_result.get_symbolized_data()
  security_flag = crash_result.is_security_issue()
  if (security_flag != testcase.security_flag or
      state.crash_state != expected_crash_state):
    logs.log_warn('Ignoring unrelated crash.\n'
                  'State: %s (expected %s)\n'
                  'Security: %s (expected %s)\n'
                  'Output: %s\n' %
                  (state.crash_state, expected_crash_state, security_flag,
                   testcase.security_flag, state.crash_stacktrace))
    return None, None

  with open(output_file_path, 'rb') as file_handle:
    minimized_keys = blobs.write_blob(file_handle)

  testcase.minimized_keys = minimized_keys
  testcase.put()

  return output_file_path, crash_result


def _update_crash_result(testcase, crash_result, command):
  """Update testcase with crash result."""
  min_state = crash_result.get_symbolized_data()
  min_unsymbolized_crash_stacktrace = crash_result.get_stacktrace(
      symbolized=False)
  min_crash_stacktrace = utils.get_crash_stacktrace_output(
      command, min_state.crash_stacktrace, min_unsymbolized_crash_stacktrace)
  testcase.crash_type = min_state.crash_type
  testcase.crash_address = min_state.crash_address
  testcase.crash_state = min_state.crash_state
  testcase.crash_stacktrace = data_handler.filter_stacktrace(
      min_crash_stacktrace)


def _skip_minimization(testcase, message, crash_result=None, command=None):
  """Skip minimization for a testcase."""
  testcase = data_handler.get_testcase_by_id(testcase.key.id())
  testcase.minimized_keys = testcase.fuzzed_keys

  if crash_result:
    _update_crash_result(testcase, crash_result, command)

  data_handler.update_testcase_comment(testcase, data_types.TaskState.ERROR,
                                       message)
  create_additional_tasks(testcase)


def do_libfuzzer_minimization(testcase, testcase_file_path):
  """Use libFuzzer's built-in minimizer where appropriate."""
  is_overriden_job = bool(environment.get_value('ORIGINAL_JOB_NAME'))

  def handle_unreproducible():
    # Be more lenient with marking testcases as unreproducible when this is a
    # job override.
    if is_overriden_job:
      _skip_minimization(testcase, 'Unreproducible on overridden job')
    else:
      task_creation.mark_unreproducible_if_flaky(testcase, True)

  timeout = environment.get_value('LIBFUZZER_MINIMIZATION_TIMEOUT', 600)
  rounds = environment.get_value('LIBFUZZER_MINIMIZATION_ROUNDS', 5)
  current_testcase_path = testcase_file_path
  last_crash_result = None

  # Get initial crash state.
  initial_crash_result = _run_libfuzzer_testcase(
      testcase, testcase_file_path, crash_retries=None)  # Use default retries.
  if not initial_crash_result.is_crash():
    logs.log_warn('Did not crash. Output:\n' +
                  initial_crash_result.get_stacktrace(symbolized=True))
    handle_unreproducible()
    return

  if testcase.security_flag != initial_crash_result.is_security_issue():
    logs.log_warn('Security flag does not match.')
    handle_unreproducible()
    return

  task_creation.mark_unreproducible_if_flaky(testcase, False)

  expected_state = initial_crash_result.get_symbolized_data()
  logs.log('Initial crash state: %s\n' % expected_state.crash_state)

  # Minimize *_OPTIONS env variable first.
  env = {}
  for tool in environment.SUPPORTED_MEMORY_TOOLS_FOR_OPTIONS:
    options_env_var = tool + '_OPTIONS'
    options = environment.get_memory_tool_options(options_env_var)
    if not options:
      continue

    minimized_options = options.copy()
    for options_name, options_value in six.iteritems(options):
      if utils.is_oss_fuzz() and options_name in MANDATORY_OSS_FUZZ_OPTIONS:
        continue

      minimized_options.pop(options_name)
      environment.set_memory_tool_options(options_env_var, minimized_options)

      reproduced = False
      for _ in range(MINIMIZE_SANITIZER_OPTIONS_RETRIES):
        crash_result = _run_libfuzzer_testcase(testcase, testcase_file_path)
        if (crash_result.is_crash() and crash_result.is_security_issue() ==
            initial_crash_result.is_security_issue() and
            crash_result.get_type() == initial_crash_result.get_type() and
            crash_result.get_state() == initial_crash_result.get_state()):
          reproduced = True
          break

      if reproduced:
        logs.log(
            'Removed unneeded {options_env_var} option: {options_name}'.format(
                options_env_var=options_env_var, options_name=options_name))
      else:
        minimized_options[options_name] = options_value
        logs.log(
            'Skipped needed {options_env_var} option: {options_name}'.format(
                options_env_var=options_env_var, options_name=options_name),
            crash_type=crash_result.get_type(),
            crash_state=crash_result.get_state(),
            security_flag=crash_result.is_security_issue())

    environment.set_memory_tool_options(options_env_var, minimized_options)
    env[options_env_var] = environment.get_memory_tool_options(options_env_var)
  if env:
    testcase = data_handler.get_testcase_by_id(testcase.key.id())
    testcase.set_metadata('env', env)

  # We attempt minimization multiple times in case one round results in an
  # incorrect state, or runs into another issue such as a slow unit.
  for round_number in range(1, rounds + 1):
    logs.log('Minimizing round %d.' % round_number)
    output_file_path, crash_result = _run_libfuzzer_tool(
        'minimize',
        testcase,
        current_testcase_path,
        timeout,
        expected_state.crash_state,
        set_dedup_flags=True)
    if output_file_path:
      last_crash_result = crash_result
      current_testcase_path = output_file_path

  if not last_crash_result:
    repro_command = testcase_manager.get_command_line_for_application(
        file_to_run=testcase_file_path, needs_http=testcase.http_flag)
    _skip_minimization(
        testcase,
        'LibFuzzer minimization failed',
        crash_result=initial_crash_result,
        command=repro_command)
    return

  logs.log('LibFuzzer minimization succeeded.')

  if utils.is_oss_fuzz():
    # Scrub the testcase of non-essential data.
    cleansed_testcase_path = do_libfuzzer_cleanse(
        testcase, current_testcase_path, expected_state.crash_state)
    if cleansed_testcase_path:
      current_testcase_path = cleansed_testcase_path

  # Finalize the test case if we were able to reproduce it.
  repro_command = testcase_manager.get_command_line_for_application(
      file_to_run=current_testcase_path, needs_http=testcase.http_flag)
  finalize_testcase(
      testcase.key.id(), repro_command, last_crash_result, flaky_stack=False)

  # Clean up after we're done.
  shell.clear_testcase_directories()


def do_libfuzzer_cleanse(testcase, testcase_file_path, expected_crash_state):
  """Cleanse testcase using libFuzzer."""
  timeout = environment.get_value('LIBFUZZER_CLEANSE_TIMEOUT', 180)
  output_file_path, _ = _run_libfuzzer_tool(
      'cleanse', testcase, testcase_file_path, timeout, expected_crash_state)

  if output_file_path:
    logs.log('LibFuzzer cleanse succeeded.')

  return output_file_path


def do_line_minimization(test_function, get_temp_file, data, deadline, threads,
                         cleanup_interval, delete_temp_files):
  """Line-by-line minimization strategy."""
  current_minimizer = delta_minimizer.DeltaMinimizer(
      test_function,
      max_threads=threads,
      deadline=deadline,
      cleanup_function=process_handler.cleanup_stale_processes,
      single_thread_cleanup_interval=cleanup_interval,
      get_temp_file=get_temp_file,
      delete_temp_files=delete_temp_files,
      progress_report_function=functools.partial(logs.log))
  return current_minimizer.minimize(data)


def do_html_minimization(test_function, get_temp_file, data, deadline, threads,
                         cleanup_interval, delete_temp_files):
  """HTML minimization strategy."""
  current_minimizer = html_minimizer.HTMLMinimizer(
      test_function,
      max_threads=threads,
      deadline=deadline,
      cleanup_function=process_handler.cleanup_stale_processes,
      single_thread_cleanup_interval=cleanup_interval,
      get_temp_file=get_temp_file,
      delete_temp_files=delete_temp_files,
      progress_report_function=functools.partial(logs.log))
  try:
    return current_minimizer.minimize(data)
  except minimizer_errors.AntlrDecodeError:
    return do_line_minimization(test_function, get_temp_file, data, deadline,
                                threads, cleanup_interval, delete_temp_files)


def minimize_file(file_path,
                  test_function,
                  get_temp_file,
                  data,
                  deadline,
                  threads,
                  cleanup_interval,
                  delete_temp_files=True):
  """Attempt to minimize a single file."""
  # Specialized minimization strategy for IPC dumps.
  if file_path.endswith(testcase_manager.IPCDUMP_EXTENSION):
    return do_ipc_dump_minimization(test_function, get_temp_file, file_path,
                                    deadline, threads, cleanup_interval,
                                    delete_temp_files)

  # Specialized minimization strategy for javascript.
  if file_path.endswith('.js'):
    return do_js_minimization(test_function, get_temp_file, data, deadline,
                              threads, cleanup_interval, delete_temp_files)

  if file_path.endswith('.html'):
    return do_html_minimization(test_function, get_temp_file, data, deadline,
                                threads, cleanup_interval, delete_temp_files)

  # We could not identify another strategy for this file, so use the default.
  return do_line_minimization(test_function, get_temp_file, data, deadline,
                              threads, cleanup_interval, delete_temp_files)



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Stack parsing constants."""

import re

from clusterfuzz._internal.crash_analysis.stack_parsing import stack_parser

C_CPP_EXTENSIONS = ['c', 'cc', 'cpp', 'cxx', 'h', 'hh', 'hpp', 'hxx']

# Patterns which cannot be compiled directly, or which are used for direct
# comparison.
CHECK_FAILURE_PATTERN = r'Check failed: '
JNI_ERROR_STRING = r'JNI DETECTED ERROR IN APPLICATION:'

# Common log prefix format for Google fatal logs.
GOOGLE_LOG_FATAL_PREFIX = r'^F\d{4}\s+\d{2}:\d{2}:\d{2}\.\d+\s+\d+\s+(.*):\d+\]'

# Compiled regular expressions.
ANDROID_ABORT_REGEX = re.compile(r'^Abort message: (.*)')
ANDROID_FATAL_EXCEPTION_REGEX = re.compile(r'.*FATAL EXCEPTION.*:')
ANDROID_KERNEL_ERROR_REGEX = re.compile(
    r'.*Internal error: (Oops)?( -|:) (BUG|[0-9a-fA-F]+)')
ANDROID_KERNEL_STACK_FRAME_REGEX = re.compile(
    # e.g. "[ 1998.156940] [<c0667574>] "
    r'[^(]*\[\<([x0-9a-fA-F]+)\>\]\s+'
    # e.g. "(msm_vidc_prepare_buf+0xa0/0x124)"; function (3), offset (4)
    r'\(?(([\w]+)\+([\w]+)/[\w]+)\)?')
ANDROID_KERNEL_TIME_REGEX = re.compile(r'^\[\s*\d+\.\d+\]\s')
# Parentheses are optional.
ANDROID_PROCESS_NAME_REGEX = re.compile(r'.*[(](.*)[)]$')
ANDROID_SEGV_REGEX = re.compile(r'.*signal.*\(SIG.*fault addr ([^ ]*)(.*)')
ASAN_INVALID_FREE_REGEX = re.compile(
    r'.*AddressSanitizer: '
    r'attempting free on address which was not malloc\(\)-ed: '
    r'([xX0-9a-fA-F]+)')
ASAN_DOUBLE_FREE_REGEX = re.compile(
    r'.*(AddressSanitizer).*double-free'
    r' on (unknown address |address |)([xX0-9a-fA-F]+)')
ASAN_MEMCPY_OVERLAP_REGEX = re.compile(
    r'.*(AddressSanitizer).*memcpy-param-overlap'
    r'[^\[]*([\[].*[)])')
ASAN_REGEX = re.compile(
    r'.*ERROR: (HWAddressSanitizer|AddressSanitizer)[: ]*[ ]*([^(:;]+)')
ASSERT_REGEX = re.compile(
    r'(?:\[.*?\]|.*\.(?:%s):.*)?' % ('|'.join(C_CPP_EXTENSIONS)) +
    r'\s*(?:ASSERT(?:ION)? FAIL(?:URE|ED)|panic): (.*)', re.IGNORECASE)
ASSERT_REGEX_GOOGLE = re.compile(GOOGLE_LOG_FATAL_PREFIX +
                                 r'.*assertion failed at\s.*\sin\s*.*: (.*)')
ASSERT_REGEX_GLIBC = re.compile(
    r'.*:\s*assertion [`\'"]?(.*?)[`\'"]? failed\.?$', re.IGNORECASE)
ASSERT_NOT_REACHED_REGEX = re.compile(r'^\s*SHOULD NEVER BE REACHED\s*$')
CFI_ERROR_REGEX = re.compile(
    r'(.*): runtime error: control flow integrity check for type (.*) '
    r'failed during (.*vtable address ([xX0-9a-fA-F]+)|.*)')
CFI_INVALID_DOWNCAST_REGEX = re.compile(r'.*note: vtable is of type (.*)')
CFI_INVALID_VPTR_REGEX = re.compile(r'.*note: invalid vtable$')
CFI_FUNC_DEFINED_HERE_REGEX = re.compile(r'.*note: .* defined here$')
CFI_NODEBUG_ERROR_MARKER_REGEX = re.compile(
    r'CFI: Most likely a control flow integrity violation;.*')
CHROME_CHECK_FAILURE_REGEX = re.compile(
    r'\s*\[[^\]]*[:]([^\](]*).*\].*Check failed[:]\s*(.*)')
CHROME_STACK_FRAME_REGEX = re.compile(
    r'[ ]*(#(?P<frame_id>[0-9]+)[ ]'  # frame id (2)
    r'([xX0-9a-fA-F]+)[ ])'  # addr (3)
    r'([^/\\]+)$')  # rest, usually fun (4); may have off
CHROME_WIN_STACK_FRAME_REGEX = re.compile(
    r'[ ]*([^/\\]+) '  # fun (1)
    r'\[([xX0-9a-fA-F]+)\+'  # fun_base (2)
    r'(\d+)\]'  # off[dec] (3)
    r'( \((.*):(\d+)\))?')  # if available, file (5) and line (6)
CHROME_MAC_STACK_FRAME_REGEX = re.compile(
    r'(?P<frame_id>\d+)\s+'  # frame id (1)
    r'(([\w ]+)|(\?\?\?))\s+'  # image (2)
    r'([xX0-9a-fA-F]+)\s+'  # addr[hex] (5)
    r'([^/\\]+)\s*\+\s*'  # fun (6)
    r'(\d+)')  # off[dec] (7)
MSAN_TSAN_REGEX = re.compile(
    r'.*(ThreadSanitizer|MemorySanitizer):\s+(?!ABRT)(?!ILL)([^(:]+)')
FATAL_ERROR_GENERIC_FAILURE = re.compile(r'#\s+()(.*)')
FATAL_ERROR_CHECK_FAILURE = re.compile(
    r'#\s+(Check failed: |RepresentationChangerError: node #\d+:)(.*)')
FATAL_ERROR_DCHECK_FAILURE = re.compile(r'#\s+(Debug check failed: )(.*)')
FATAL_ERROR_REGEX = re.compile(r'#\s*Fatal error in (.*)')
FATAL_ERROR_LINE_REGEX = re.compile(r'#\s*Fatal error in (.*), line [0-9]+')
FATAL_ERROR_UNREACHABLE = re.compile(r'# un(reachable|implemented) code')
GENERIC_SEGV_HANDLER_REGEX = re.compile(
    'Received signal 11 SEGV_[A-Z]+ ([0-9a-f]*)')
GOOGLE_CHECK_FAILURE_REGEX = re.compile(GOOGLE_LOG_FATAL_PREFIX +
                                        r'\s*Check failed[:]\s*(.*)')
GOOGLE_LOG_FATAL_REGEX = re.compile(GOOGLE_LOG_FATAL_PREFIX + r'\s*(.*)')
HWASAN_ALLOCATION_TAIL_OVERWRITTEN_ADDRESS_REGEX = re.compile(
    r'.*ERROR: HWAddressSanitizer: allocation-tail-overwritten; '
    r'heap object \[([xX0-9a-fA-F]+),.*of size')
JAZZER_JAVA_EXCEPTION_REGEX = re.compile('== Java Exception: .*')
JAVA_EXCEPTION_CRASH_STATE_REGEX = re.compile(r'\s*at (.*)\(.*\)')
KASAN_ACCESS_TYPE_REGEX = re.compile(r'(Read|Write) of size ([0-9]+)')
KASAN_ACCESS_TYPE_ADDRESS_REGEX = re.compile(
    r'(Read|Write) of size ([0-9]+) at (addr|address) ([a-f0-9]+)')
KASAN_CRASH_TYPE_ADDRESS_REGEX = re.compile(
    r'BUG: KASAN: (.*) (in|on).*(addr|address) ([a-f0-9]+)')
KASAN_CRASH_TYPE_FUNCTION_REGEX = re.compile(
    r'BUG: KASAN: (.*) (in|on).* ([\w]+)\+([\w]+)/([\w]+)')
KASAN_GPF_REGEX = re.compile(r'general protection fault:.*KASAN')
LIBFUZZER_DEADLY_SIGNAL_REGEX = re.compile(
    r'.*ERROR:\s*libFuzzer:\s*deadly signal')
LIBFUZZER_FUZZ_TARGET_EXITED_REGEX = re.compile(
    r'.*ERROR:\s*libFuzzer:\s*fuzz target exited')
LIBFUZZER_OVERWRITES_CONST_INPUT_REGEX = re.compile(
    r'.*ERROR:\s*libFuzzer:\s*fuzz target overwrites its const input')
LIBFUZZER_TIMEOUT_REGEX = re.compile(r'.*ERROR:\s*libFuzzer:\s*timeout')
LIBRARY_NOT_FOUND_ANDROID_REGEX = re.compile(
    r'.*: library ([`\'"])(.*)\1 not found')
LIBRARY_NOT_FOUND_LINUX_REGEX = re.compile(
    r'.*error while loading shared libraries: ([^:]*): '
    r'cannot open shared object file')
LINUX_GDB_CRASH_TYPE_REGEX = re.compile(r'Program received signal ([a-zA-Z]+),')
LINUX_GDB_CRASH_ADDRESS_REGEX = re.compile(r'rip[ ]+([xX0-9a-fA-F]+)')
LINUX_GDB_CRASH_ADDRESS_NO_REGISTERS_REGEX = re.compile(
    r'^(0[xX][0-9a-fA-F]+)\s+in\s+')
LSAN_DIRECT_LEAK_REGEX = re.compile(r'Direct leak of ')
LSAN_INDIRECT_LEAK_REGEX = re.compile(r'Indirect leak of ')
MAC_GDB_CRASH_ADDRESS_REGEX = re.compile(
    r'Reason:.*at address[^0-9]*([0-9a-zA-Z]+)')
OUT_OF_MEMORY_REGEX = re.compile(r'.*(?:%s).*' % '|'.join([
    r'# Allocation failed.*out of memory',
    r'::OnNoMemory',
    r'ERROR.*Sanitizer failed to allocate',
    r'FatalProcessOutOfMemory',  # V8
    r'Fatal (?:process|JavaScript) out of memory:',  # V8
    r'Fatal JavaScript invalid size error',  # V8
    r'FX_OutOfMemoryTerminate',
    r'Out of memory\. Dying.',
    r'Out of memory\. size=',
    r'Sanitizer: allocation-size-too-big',
    r'Sanitizer: calloc-overflow',
    r'Sanitizer: calloc parameters overflow',
    r'Sanitizer: requested allocation size.*exceeds maximum supported size',
    r'TerminateBecauseOutOfMemory',
    r'allocator is out of memory trying to allocate',
    r'blinkGCOutOfMemory',
    r'couldnt allocate.*Out of memory',
    r'libFuzzer: out-of-memory \(',
    r'rss limit exhausted',
    r'in rust_oom'
]))
RUNTIME_ERROR_REGEX = re.compile(r'#\s*Runtime error in (.*)')
RUNTIME_ERROR_LINE_REGEX = re.compile(r'#\s*Runtime error in (.*), line [0-9]+')
RUST_ASSERT_REGEX = re.compile(r'thread\s.*\spanicked at \'([^\']*)',
                               re.IGNORECASE)
SAN_ABRT_REGEX = re.compile(r'.*[a-zA-Z]+Sanitizer: ABRT ')
SAN_BREAKPOINT_REGEX = re.compile(r'.*[a-zA-Z]+Sanitizer: breakpoint ')
SAN_CHECK_FAILURE_REGEX = re.compile(
    r'.*Sanitizer CHECK failed[:]\s*[^ ]*\s*(.*)')
SAN_CRASH_TYPE_ADDRESS_REGEX = re.compile(
    r'[ ]*([^ ]*|Atomic [^ ]*) of size ([^ ]*) at ([^ ]*)')
SAN_DEADLYSIGNAL_REGEX = re.compile(r'.*:DEADLYSIGNAL')
SAN_FPE_REGEX = re.compile(r'.*[a-zA-Z]+Sanitizer: FPE ')
SAN_ILL_REGEX = re.compile(r'.*[a-zA-Z]+Sanitizer: ILL ')
SAN_TRAP_REGEX = re.compile(r'.*[a-zA-Z]+Sanitizer: TRAP ')
SAN_SEGV_CRASH_TYPE_REGEX = re.compile(
    r'.*The signal is caused by a ([A-Z]+) memory access.')
# FIXME: Replace when better ways to check signal crashes are available.
SAN_SIGNAL_REGEX = re.compile(r'.*SCARINESS: (\d+) \(signal\)', re.DOTALL)
SAN_STACK_FRAME_REGEX = re.compile(
    # frame id (1)
    r'\s*#(?P<frame_id>\d+)\s+'
    # addr (2)
    r'([xX0-9a-fA-F]+)\s+'
    # Format is [in {fun}[+{off}]] [{file}[:{line}[:{char}]]] [({mod}[+{off}])]
    # If there is fun and mod/file info, extract
    # fun+off, where fun (7, 5, 23), off (8)
    r'((in\s*(((.*)\+([xX0-9a-fA-F]+))|(.*)) '
    r'('
    # file:line:char, where file (12, 16), line (13, 17), char (14)
    r'(([^ ]+):(\d+):(\d+))|(([^ ]+):(\d+))'
    # or mod+off, where mod (19, 31), off (21, 32)
    r'|'
    r'(\(([^+]+)(\+([xX0-9a-fA-F]+))?\)))'
    r')'
    # If there is only fun info, extract
    r'|'
    r'(in\s*(((.*)\+([xX0-9a-fA-F]+))|(.*)))'
    # If there is only mod info, extract
    r'|'
    r'(\((((.*)\+([xX0-9a-fA-F]+))|(.*))\))'
    r')')
SAN_ADDR_REGEX = re.compile(r'.*(ERROR: [a-zA-Z]+Sanitizer)[: ]*(.*) on '
                            r'(unknown address |address |)([xX0-9a-fA-F]+)')
SAN_SEGV_REGEX = re.compile(r'.*([a-zA-Z]+Sanitizer).*(SEGV|access-violation) '
                            r'on unknown address ([xX0-9a-fA-F]+)')
SECURITY_CHECK_FAILURE_REGEX = re.compile(
    r'.*\[[^\]]*[:]([^\](]*).*\].*Security CHECK failed[:]\s*(.*)\.\s*')
SECURITY_DCHECK_FAILURE_REGEX = re.compile(
    r'.*\[[^\]]*[:]([^\](]*).*\].*Security DCHECK failed[:]\s*(.*)\.\s*')
UBSAN_DIVISION_BY_ZERO_REGEX = re.compile(r'.*division by zero.*')
UBSAN_FLOAT_CAST_OVERFLOW_REGEX = re.compile(r'.*outside the range of '
                                             r'representable values.*')
UBSAN_INCORRECT_FUNCTION_POINTER_REGEX = re.compile(
    r'.*call to function [^\s]+ through pointer to incorrect function type.*')
UBSAN_INDEX_OOB_REGEX = re.compile(r'.*out of bounds for type.*')
UBSAN_UNSIGNED_INTEGER_OVERFLOW_REGEX = re.compile(
    r'.*unsigned integer overflow.*')
UBSAN_INTEGER_OVERFLOW_REGEX = re.compile(
    r'.*(integer overflow|'
    r'(negation|division) of.*cannot be represented in type).*')
UBSAN_INVALID_BOOL_VALUE_REGEX = re.compile(
    r'.*not a valid value for type \'(bool|BOOL)\'.*')
UBSAN_INVALID_BUILTIN_REGEX = re.compile(r'.*, which is not a valid argument.*')
UBSAN_INVALID_ENUM_VALUE_REGEX = re.compile(r'.*not a valid value for type.*')
UBSAN_MISALIGNED_ADDRESS_REGEX = re.compile(r'.*misaligned address.*')
UBSAN_NO_RETURN_VALUE_REGEX = re.compile(
    r'.*reached the end of a value-returning function.*')
UBSAN_NULL_ARGUMENT_REGEX = re.compile(
    r'.*null pointer passed as .*, which is declared to never be null.*')
UBSAN_NULL_POINTER_READ_REGEX = re.compile(r'.*load of null pointer.*')
UBSAN_NULL_POINTER_REFERENCE_REGEX = re.compile(
    r'.*(binding to|access within|call on) null pointer.*')
UBSAN_NULL_POINTER_WRITE_REGEX = re.compile(r'.*store to null pointer.*')
UBSAN_OBJECT_SIZE_REGEX = re.compile(
    r'.*address .* with insufficient space for an object of type.*')
UBSAN_POINTER_OVERFLOW_REGEX = re.compile(
    r'.*((addition|subtraction) of unsigned offset |'
    r'pointer index expression with base |'
    r'applying non-zero offset [0-9]+ to null pointer|'
    r'applying zero offset to null pointer).*')
UBSAN_RETURNS_NONNULL_ATTRIBUTE_REGEX = re.compile(
    r'.*null pointer returned from function declared to never return null.*')
UBSAN_RUNTIME_ERROR_REGEX = re.compile(r'(.*): runtime error: (.*)')
UBSAN_SHIFT_ERROR_REGEX = re.compile(r'.*shift.*')
UBSAN_UNREACHABLE_REGEX = re.compile(
    r'.*execution reached an unreachable program point.*')
UBSAN_VLA_BOUND_REGEX = re.compile(
    r'.*variable length array bound evaluates to non-positive value.*')
UBSAN_VPTR_REGEX = re.compile(
    r'(.*): runtime error: '
    r'(member access within|member call on|downcast of)'
    r' address ([xX0-9a-fA-F]+) .* of type (.*)')
UBSAN_VPTR_INVALID_DOWNCAST_REGEX = re.compile(
    r'.*note: object is of type (.*)')
UBSAN_VPTR_INVALID_OFFSET_REGEX = re.compile(
    r'.*at offset (\d+) within object of type (.*)')
UBSAN_VPTR_INVALID_VPTR_REGEX = re.compile(r'.*note: object has invalid vptr')
V8_ABORT_FAILURE_REGEX = re.compile(r'^abort: (CSA_ASSERT failed:.*)')
V8_ABORT_METADATA_REGEX = re.compile(r'(.*) \[(.*):\d+\]$')
V8_CORRECTNESS_FAILURE_REGEX = re.compile(r'#\s*V8 correctness failure')
V8_CORRECTNESS_METADATA_REGEX = re.compile(
    r'#\s*V8 correctness ((configs|sources|suppression): .*)')
WINDOWS_CDB_STACK_FRAME_REGEX = re.compile(
    r'([0-9a-zA-Z`]+) '  # Child EBP or SP; remove ` if needed (1)
    r'([0-9a-zA-Z`]+) '  # RetAddr; remove ` if needed (2)
    r'([0-9a-zA-Z_]+)'  # mod (3)
    r'!(.*)\+'  # fun (4)
    r'([xX0-9a-fA-F]+)')  # off (5)
WINDOWS_CDB_STACK_START_REGEX = re.compile(r'ChildEBP RetAddr')
WINDOWS_CDB_CRASH_TYPE_ADDRESS_REGEX = re.compile(
    r'Attempt to (.*) [^ ]* address (.*)')
WINDOWS_CDB_CRASH_TYPE_REGEX = re.compile(
    r'.*DEFAULT_BUCKET_ID[ ]*[:][ ]*([a-zA-Z_]+)')
WINDOWS_CDB_STACK_OVERFLOW_REGEX = re.compile(
    r'.*ExceptionCode: .*\(Stack overflow\).*')

# Golang specific regular expressions.
GOLANG_DIVISION_BY_ZERO_REGEX = re.compile(
    r'^panic: runtime error: integer divide by zero.*')
GOLANG_INDEX_OUT_OF_RANGE_REGEX = re.compile(
    r'^panic: runtime error: index out of range.*')
GOLANG_INVALID_MEMORY_ADDRESS_REGEX = re.compile(
    r'^panic: runtime error: invalid memory address.*')
GOLANG_MAKESLICE_LEN_OUT_OF_RANGE_REGEX = re.compile(
    r'^panic: runtime error: makeslice: len out of range.*')
GOLANG_SLICE_BOUNDS_OUT_OF_RANGE_REGEX = re.compile(
    r'^panic: runtime error: slice bounds out of range.*')
GOLANG_STACK_OVERFLOW_REGEX = re.compile(r'^fatal error: stack overflow.*')

GOLANG_CRASH_TYPES_MAP = [
    (GOLANG_DIVISION_BY_ZERO_REGEX, 'Integer divide by zero'),
    (GOLANG_INDEX_OUT_OF_RANGE_REGEX, 'Index out of range'),
    (GOLANG_INVALID_MEMORY_ADDRESS_REGEX, 'Invalid memory address'),
    (GOLANG_MAKESLICE_LEN_OUT_OF_RANGE_REGEX, 'Makeslice: len out of range'),
    (GOLANG_SLICE_BOUNDS_OUT_OF_RANGE_REGEX, 'Slice bounds out of range'),
    (GOLANG_STACK_OVERFLOW_REGEX, 'Stack overflow'),
]

GOLANG_FATAL_ERROR_REGEX = re.compile(r'^fatal error: (.*)')

GOLANG_STACK_FRAME_FUNCTION_REGEX = re.compile(
    r'^([0-9a-zA-Z\.\-\_\\\/\(\)\*]+)\([x0-9a-f\s,\.]*\)$')

# Python specific regular expressions.
PYTHON_UNHANDLED_EXCEPTION = re.compile(
    r'^\s*=== Uncaught Python exception: ===$')

PYTHON_CRASH_TYPES_MAP = [
    (PYTHON_UNHANDLED_EXCEPTION, 'Uncaught exception'),
]

PYTHON_STACK_FRAME_FUNCTION_REGEX = re.compile(
    #  File "<embedded stdlib>/gzip.py", line 421, in _read_gzip_header
    r'^\s*File "([^"]+)", line (\d+), in (.+)$')

# Mappings of Android kernel error status codes to strings.
ANDROID_KERNEL_STATUS_TO_STRING = {
    0b0001: 'Alignment Fault',
    0b0100: 'Instruction Cache Maintenance Fault',
    0b1100: 'L1 Translation',
    0b1110: 'L2 Translation',
    0b0101: 'Translation Fault, Section',
    0b0111: 'Translation Fault, Page',
    0b0011: 'Access Flag Fault, Section',
    0b0110: 'Access Flag Fault, Page',
    0b1001: 'Domain Fault, Section',
    0b1011: 'Domain Fault, Page',
    0b1101: 'Permission Fault, Section',
    0b1111: 'Permissions Fault, Page',
}

# Ignore lists.
STACK_FRAME_IGNORE_REGEXES = [
    # Function names (exact match).
    r'^abort$',
    r'^exit$',
    r'^pthread_create$',
    r'^pthread_kill$',
    r'^raise$',
    r'^tgkill$',
    r'^__chk_fail$',
    r'^__fortify_fail$',

    # Function names (startswith).
    r'^(|__)aeabi_',
    r'^(|__)memcmp',
    r'^(|__)memcpy',
    r'^(|__)memmove',
    r'^(|__)memset',
    r'^(|__)strcmp',
    r'^(|__)strcpy',
    r'^(|__)strdup',
    r'^(|__)strlen',
    r'^(|__)strncpy',
    r'^<null>',
    r'^Abort\(',
    r'^CFCrash',
    r'^ExitCallback',
    r'^IsSandboxedProcess',
    r'^LLVMFuzzerTestOneInput',
    r'^MSanAtExitWrapper',
    r'^New',
    r'^RaiseException',
    r'^SbSystemBreakIntoDebugger',
    r'^SignalAction',
    r'^SignalHandler',
    r'^TestOneProtoInput',
    r'^V8_Fatal',
    r'^WTF::',
    r'^WTFCrash',
    r'^X11Error',
    r'^_L_unlock_',
    r'^_\$LT\$',
    r'^__GI_',
    r'^__asan::',
    r'^__asan_',
    r'^__assert_',
    r'^__cxa_atexit',
    r'^__cxa_rethrow',
    r'^__cxa_throw',
    r'^__dump_stack',
    r'^__hwasan::',
    r'^__hwasan_',
    r'^__interceptor_',
    r'^__kasan_',
    r'^__libc_',
    r'^__lsan::',
    r'^__lsan_',
    r'^__msan::',
    r'^__msan_',
    r'^__pthread_kill',
    r'^__run_exit_handlers',
    r'^__rust_try',
    r'^__sanitizer::',
    r'^__sanitizer_',
    r'^__tsan::',
    r'^__tsan_',
    r'^__ubsan::',
    r'^__ubsan_',
    r'^_asan_',
    r'^_hwasan_',
    r'^_lsan_',
    r'^_msan_',
    r'^_objc_terminate',
    r'^_sanitizer_',
    r'^_start',
    r'^_tsan_',
    r'^_ubsan_',
    r'^abort',
    r'^alloc::',
    r'^android\.app\.ActivityManagerProxy\.',
    r'^android\.os\.Parcel\.',
    r'^art::Thread::CreateNativeThread',
    r'^asan_',
    r'^asan\.module_ctor',
    r'^asan\.module_dtor',
    r'^calloc',
    r'^check_memory_region',
    r'^common_exit',
    r'^delete',
    r'^demangling_terminate_handler',
    r'^dump_backtrace',
    r'^dump_stack',
    r'^exit_or_terminate_process',
    r'^fpehandler\(',
    r'^free',
    r'^fuzzer::',
    r'^g_log',
    r'^generic_cpp_',
    r'^gsignal',
    r'^kasan_',
    r'^libfuzzer_sys::initialize',
    r'^main',
    r'^malloc',
    r'^mozalloc_',
    r'^new',
    r'^object_err',
    r'^operator',
    r'^print_trailer',
    r'^realloc',
    r'^rust_begin_unwind',
    r'^rust_fuzzer_test_input',
    r'^rust_oom',
    r'^scanf',
    r'^show_stack',
    r'^std::__terminate',
    r'^std::panic',
    r'^std::process::abort',
    r'^std::sys::unix::abort',
    r'^__scrt_common_main_seh',

    # Functions names (contains).
    r'.*ASAN_OnSIGSEGV',
    r'.*BaseThreadInitThunk',
    r'.*DebugBreak',
    r'.*DefaultDcheckHandler',
    r'.*ForceCrashOnSigAbort',
    r'.*MemoryProtection::CMemoryProtector',
    r'.*PartitionAlloc',
    r'.*RtlFreeHeap',
    r'.*RtlInitializeExceptionChain',
    r'.*RtlReportCriticalFailure',
    r'.*RtlUserThreadStart',
    r'.*RtlpHeapHandleError',
    r'.*RtlpLogHeapFailure',
    r'.*SkDebugf',
    r'.*StackDumpSignalHandler',
    r'.*__android_log_assert',
    r'.*__tmainCRTStartup',
    r'.*_asan_rtl_',
    r'.*agent::asan::',
    r'.*allocator_shim',
    r'.*asan_Heap',
    r'.*asan_check_access',
    r'.*asan_osx_dynamic\.dylib',
    r'.*assert',
    r'.*base::FuzzedDataProvider',
    r'.*base::allocator',
    r'.*base::android::CheckException',
    r'.*base::debug::BreakDebugger',
    r'.*base::debug::CollectStackTrace',
    r'.*base::debug::StackTrace::StackTrace',
    r'.*ieee754\-',
    r'.*libpthread',
    r'.*logger',
    r'.*logging::CheckError',
    r'.*logging::ErrnoLogMessage',
    r'.*logging::LogMessage',
    r'.*stdext::exception::what',
    r'.*v8::base::OS::Abort',

    # File paths.
    r'.* base/callback',
    r'.* /rust(|c)/',
    r'.*/AOSP\-toolchain/',
    r'.*/bindings/ToV8\.h',
    r'.*/crosstool/',
    r'.*/gcc/',
    r'.*/glibc\-',
    r'.*/jemalloc/',
    r'.*/libc\+\+',
    r'.*/libc/',
    r'.*/llvm\-build/',
    r'.*/minkernel/crts/',
    r'.*/sanitizer_common/',
    r'.*/tcmalloc/',
    r'.*/vc/include/',
    r'.*/vctools/crt/',
    r'.*/win_toolchain/',
    r'.*libc\+\+/',

    # Wrappers from honggfuzz/libhfuzz/memorycmp.c.
    r'.*/memorycmp\.c',

    # Others (uncategorized).
    r'.*\+Unknown',
    r'.*<unknown module>',
    r'.*Inline Function @',
    r'^<unknown>$',
    r'^\[vdso\]$',

    # Golang specific frames to ignore.
    r'^panic$',
    r'^runtime\.',

    # Fuchsia specific.
    r'^CrashTrampolineAsm',
    r'^libc_io_functions_not_implemented_use_fdio_instead',
    r'^<libclang_rt.asan.so>',
    r'^__zx_panic',
    r'^syslog::LogMessage',

    # Android kernel stack frame ignores.
    r'^print_address_description',
    r'^_etext',

    # Swift specific.
    r'^_swift_stdlib_'
]

STACK_FRAME_IGNORE_REGEXES_IF_SYMBOLIZED = [
    r'.*libc\.so',
    r'.*libc\+\+\.so',
    r'.*libc\+\+_shared\.so',
    r'.*libstdc\+\+\.so',
    r'.*libc-.*\.so',
]

IGNORE_CRASH_TYPES_FOR_ABRT_BREAKPOINT_AND_ILLS = [
    'ASSERT',
    'CHECK failure',
    'DCHECK failure',
    'Fatal error',
    'Security CHECK failure',
    'Security DCHECK failure',
]

STATE_STOP_MARKERS = [
    'Direct leak of',
    'Uninitialized value was stored to memory at',
    'allocated by thread',
    'created by main thread at',
    'located in stack of thread',
    'previously allocated by',
]

UBSAN_CRASH_TYPES_MAP = [
    (UBSAN_DIVISION_BY_ZERO_REGEX, 'Divide-by-zero'),
    (UBSAN_FLOAT_CAST_OVERFLOW_REGEX, 'Float-cast-overflow'),
    (UBSAN_INCORRECT_FUNCTION_POINTER_REGEX, 'Incorrect-function-pointer-type'),
    (UBSAN_INDEX_OOB_REGEX, 'Index-out-of-bounds'),
    (UBSAN_INVALID_BOOL_VALUE_REGEX, 'Invalid-bool-value'),
    (UBSAN_INVALID_BUILTIN_REGEX, 'Invalid-builtin-use'),
    (UBSAN_MISALIGNED_ADDRESS_REGEX, 'Misaligned-address'),
    (UBSAN_NO_RETURN_VALUE_REGEX, 'No-return-value'),
    (UBSAN_NULL_ARGUMENT_REGEX, 'Invalid-null-argument'),
    (UBSAN_NULL_POINTER_READ_REGEX, 'Null-dereference READ'),
    (UBSAN_NULL_POINTER_REFERENCE_REGEX, 'Null-dereference'),
    (UBSAN_NULL_POINTER_WRITE_REGEX, 'Null-dereference WRITE'),
    (UBSAN_OBJECT_SIZE_REGEX, 'Object-size'),
    (UBSAN_POINTER_OVERFLOW_REGEX, 'Pointer-overflow'),
    (UBSAN_RETURNS_NONNULL_ATTRIBUTE_REGEX, 'Invalid-null-return'),
    (UBSAN_SHIFT_ERROR_REGEX, 'Undefined-shift'),
    (UBSAN_UNREACHABLE_REGEX, 'Unreachable code'),
    (UBSAN_UNSIGNED_INTEGER_OVERFLOW_REGEX, 'Unsigned-integer-overflow'),
    (UBSAN_VLA_BOUND_REGEX, 'Non-positive-vla-bound-value'),

    # The following types are supersets of other types, and should be placed
    # at the end to avoid subsuming crashes from the more specialized types.
    (UBSAN_INVALID_ENUM_VALUE_REGEX, 'Invalid-enum-value'),
    (UBSAN_INTEGER_OVERFLOW_REGEX, 'Integer-overflow'),
]

# Additional regexes for cleaning up format.
STRIP_STRUCTURE_REGEXES = [
    re.compile(r'^in (.*)'),  # sanitizers have prefix for function if present
    re.compile(r'^\((.*)\)$'),  # sanitizers wrap module if no function
]

# Other constants.
LINE_LENGTH_CAP = 80
MAX_CRASH_STATE_FRAMES = 3
MAX_CYCLE_LENGTH = 10
REPEATED_CYCLE_COUNT = 3

# Stackframe format specifications.
CHROME_STACK_FRAME_SPEC = stack_parser.StackFrameSpec(
    address=3, function_name=4)
CHROME_WIN_STACK_FRAME_SPEC = stack_parser.StackFrameSpec(
    function_name=1,
    function_base=2,
    function_offset=3,
    filename=5,
    fileline=6,
    base=10)
CHROME_MAC_STACK_FRAME_SPEC = stack_parser.StackFrameSpec(
    address=5, function_name=6, function_offset=7, module_name=2, base=10)
SAN_STACK_FRAME_SPEC = stack_parser.StackFrameSpec(
    address=2,
    function_name=[7, 5, 23],
    function_offset=8,
    filename=[12, 16],
    fileline=[13, 17],
    module_name=[19, 31],
    module_offset=[21, 32])
WINDOWS_CDB_STACK_FRAME_SPEC = stack_parser.StackFrameSpec(
    address=1, function_name=4, function_offset=5, module_name=3)



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for load_bigquery_stats."""
import datetime
import unittest

import flask
import mock
import webtest

from clusterfuzz._internal.datastore import data_types
from clusterfuzz._internal.tests.test_libs import helpers as test_helpers
from clusterfuzz._internal.tests.test_libs import test_utils
from handlers.cron import load_bigquery_stats


@test_utils.with_cloud_emulators('datastore')
class LoadBigQueryStatsTest(unittest.TestCase):
  """Test LoadBigQueryStatsTest."""

  def setUp(self):
    flaskapp = flask.Flask('testflask')
    flaskapp.add_url_rule(
        '/load-bigquery-stats',
        view_func=load_bigquery_stats.Handler.as_view('/load-bigquery-stats'))
    self.app = webtest.TestApp(flaskapp)

    data_types.Fuzzer(name='fuzzer', jobs=['job']).put()
    data_types.Job(name='job').put()

    test_helpers.patch(self, [
        'clusterfuzz._internal.google_cloud_utils.big_query.get_api_client',
        'handlers.base_handler.Handler.is_cron',
        'handlers.cron.load_bigquery_stats.Handler._utc_now',
    ])

    self.mock._utc_now.return_value = datetime.datetime(2016, 9, 8)  # pylint: disable=protected-access
    self.mock_bigquery = mock.MagicMock()
    self.mock.get_api_client.return_value = self.mock_bigquery

  def test_execute(self):
    """Tests executing of cron job."""
    self.app.get('/load-bigquery-stats')

    self.mock_bigquery.datasets().insert.assert_has_calls([
        mock.call(
            projectId='test-clusterfuzz',
            body={
                'datasetReference': {
                    'projectId': 'test-clusterfuzz',
                    'datasetId': 'fuzzer_stats'
                }
            }),
        mock.call().execute()
    ])

    self.mock_bigquery.tables().insert.assert_has_calls([
        mock.call(
            body={
                'timePartitioning': {
                    'type': 'DAY'
                },
                'tableReference': {
                    'projectId': 'test-clusterfuzz',
                    'tableId': 'JobRun',
                    'datasetId': 'fuzzer_stats',
                },
            },
            datasetId='fuzzer_stats',
            projectId='test-clusterfuzz'),
        mock.call().execute(),
        mock.call(
            body={
                'timePartitioning': {
                    'type': 'DAY'
                },
                'tableReference': {
                    'projectId': 'test-clusterfuzz',
                    'tableId': 'TestcaseRun',
                    'datasetId': 'fuzzer_stats',
                },
            },
            datasetId='fuzzer_stats',
            projectId='test-clusterfuzz'),
        mock.call().execute(),
    ])

    self.mock_bigquery.jobs().insert.assert_has_calls(
        [
            mock.call(
                body={
                    'configuration': {
                        'load': {
                            'destinationTable': {
                                'projectId': 'test-clusterfuzz',
                                'tableId': 'JobRun$20160907',
                                'datasetId': 'fuzzer_stats'
                            },
                            'schemaUpdateOptions': ['ALLOW_FIELD_ADDITION',],
                            'writeDisposition':
                                'WRITE_TRUNCATE',
                            'sourceUris': [
                                'gs://test-bigquery-bucket/fuzzer/JobRun/date/'
                                '20160907/*.json'
                            ],
                            'sourceFormat':
                                'NEWLINE_DELIMITED_JSON',
                            'schema': {
                                'fields': [{
                                    'type': 'INTEGER',
                                    'name': 'testcases_executed',
                                    'mode': 'NULLABLE'
                                }, {
                                    'type': 'INTEGER',
                                    'name': 'build_revision',
                                    'mode': 'NULLABLE'
                                }, {
                                    'type': 'INTEGER',
                                    'name': 'new_crashes',
                                    'mode': 'NULLABLE'
                                }, {
                                    'type': 'STRING',
                                    'name': 'job',
                                    'mode': 'NULLABLE'
                                }, {
                                    'type': 'FLOAT',
                                    'name': 'timestamp',
                                    'mode': 'NULLABLE'
                                }, {
                                    'fields': [{
                                        'type': 'STRING',
                                        'name': 'crash_type',
                                        'mode': 'NULLABLE'
                                    }, {
                                        'type': 'BOOLEAN',
                                        'name': 'is_new',
                                        'mode': 'NULLABLE'
                                    }, {
                                        'type': 'STRING',
                                        'name': 'crash_state',
                                        'mode': 'NULLABLE'
                                    }, {
                                        'type': 'BOOLEAN',
                                        'name': 'security_flag',
                                        'mode': 'NULLABLE'
                                    }, {
                                        'type': 'INTEGER',
                                        'name': 'count',
                                        'mode': 'NULLABLE'
                                    }],
                                    'type':
                                        'RECORD',
                                    'name':
                                        'crashes',
                                    'mode':
                                        'REPEATED'
                                }, {
                                    'type': 'INTEGER',
                                    'name': 'known_crashes',
                                    'mode': 'NULLABLE'
                                }, {
                                    'type': 'STRING',
                                    'name': 'fuzzer',
                                    'mode': 'NULLABLE'
                                }, {
                                    'type': 'STRING',
                                    'name': 'kind',
                                    'mode': 'NULLABLE'
                                }]
                            },
                        }
                    }
                },
                projectId='test-clusterfuzz'),
            mock.call().execute(),
            mock.call(
                body={
                    'configuration': {
                        'load': {
                            'destinationTable': {
                                'projectId': 'test-clusterfuzz',
                                'tableId': 'TestcaseRun$20160907',
                                'datasetId': 'fuzzer_stats'
                            },
                            'schemaUpdateOptions': ['ALLOW_FIELD_ADDITION',],
                            'writeDisposition':
                                'WRITE_TRUNCATE',
                            'sourceUris': [
                                'gs://test-bigquery-bucket/fuzzer/TestcaseRun/'
                                'date/20160907/*.json'
                            ],
                            'sourceFormat':
                                'NEWLINE_DELIMITED_JSON',
                        }
                    }
                },
                projectId='test-clusterfuzz'),
            mock.call().execute(),
        ],
        # Otherwise we need to mock two calls to mock.call().execute().__str__()
        # which does not seem to work well.
        any_order=True)



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Find CLs suspected of introducing regressions."""

import json
import re

import six

from clusterfuzz._internal.build_management import build_manager
from clusterfuzz._internal.build_management import revisions
from clusterfuzz._internal.config import db_config
from clusterfuzz._internal.datastore import data_handler
from clusterfuzz._internal.datastore import data_types
from clusterfuzz._internal.google_cloud_utils import pubsub
from clusterfuzz._internal.metrics import logs
from clusterfuzz._internal.system import environment

SRC_COMPONENT_OVERRIDES = {
    'https://chromium.googlesource.com/v8/v8': 'v8',
}

UNINTERESTING_LINES_REGEX = re.compile(r'.*:(VERBOSE\d|INFO|WARNING):')


def _compute_rolls(start_revisions_dict, end_revisions_dict):
  """Compute rolls between the start and end revision."""
  result = []
  for path, entry in six.iteritems(end_revisions_dict):
    url, end_sha = _extract_url_and_sha_from_deps_entry(entry)
    start_sha = None
    if path in start_revisions_dict:
      start_entry = start_revisions_dict[path]
      _, start_sha = _extract_url_and_sha_from_deps_entry(start_entry)

    # Skip adding dependencies that were unchanged between the two DEPS files.
    if start_sha == end_sha:
      continue

    current_roll = {
        'dep_path': path,
        'repo_url': url,
        'new_revision': end_sha,
    }

    # Unless this is new code, include the earlier revision as well.
    if start_sha:
      current_roll['old_revision'] = start_sha

    result.append(current_roll)

  return result


def _extract_url_and_sha_from_deps_entry(entry):
  """Split a DEPS file entry into a URL and git sha."""
  assert 'url' in entry and 'rev' in entry, 'Unexpected format: %s' % entry
  url = entry['url']
  sha = entry['rev']

  # Strip unnecessary ".git" from the URL where applicable.
  if url.endswith('.git'):
    url = url[:-len('.git')]

  return url, sha


def _format_component_revisions_for_predator(component_revisions):
  """Convert a dict of dependency rolls to the format Predator expects."""
  result = []
  for path, entry in six.iteritems(component_revisions):
    url, sha = _extract_url_and_sha_from_deps_entry(entry)
    result.append({
        'dep_path': path,
        'repo_url': url,
        'revision': sha,
    })

  return result


def _is_predator_testcase(testcase):
  """Return bool and error message for whether this testcase is applicable to
  predator or not."""
  if build_manager.is_custom_binary():
    return False, 'Not applicable to custom binaries.'

  if testcase.regression != 'NA':
    if not testcase.regression:
      return False, 'No regression range, wait for regression task to finish.'

    if ':' not in testcase.regression:
      return False, 'Invalid regression range %s.' % testcase.regression

  return True, None


def _filter_stacktrace(stacktrace):
  """Reduces noise from stacktrace and limit its size to avoid pubsub request
  limit of one megabyte."""
  filtered_stacktrace_size = 0
  filtered_stacktrace_lines = []

  for line in reversed(stacktrace.splitlines()):
    # Exclude uninteresting lines such as ones from verbose logging, info, etc.
    if UNINTERESTING_LINES_REGEX.match(line):
      continue

    new_size = filtered_stacktrace_size + len(line) + 1
    if new_size > data_types.PUBSUB_REQUEST_LIMIT:
      break

    filtered_stacktrace_lines.append(line)
    filtered_stacktrace_size = new_size

  return '\n'.join(reversed(filtered_stacktrace_lines))


def _prepare_component_revisions_dict(revision, job_type):
  """Get a component revisions dict and git sha for a revision and job type.

  Revision is expected to be a commit position."""
  revisions_dict = revisions.get_component_revisions_dict(revision, job_type)
  if not revisions_dict:
    return revisions_dict, None

  # Other code depends on the "/" prefix, but it doesn't match the DEPS format
  # that we would usually expect. Clean these values up before sending to
  # predator.
  revisions_dict['src'] = revisions_dict.pop('/src')
  return revisions_dict, revisions_dict['src']['rev']


def _set_predator_result_with_error(testcase, error_message):
  """Sets predator result with error."""
  predator_result = {
      'result': {
          'found': False,
          'project': '',
          'suspected_components': '',
          'suspected_cls': '',
          'feedback_url': '',
          'error_message': error_message,
      }
  }

  testcase = data_handler.get_testcase_by_id(testcase.key.id())
  testcase.set_metadata(
      'predator_result', predator_result, update_testcase=False)
  testcase.delete_metadata('blame_pending', update_testcase=False)
  testcase.put()


def _prepare_predator_message(testcase):
  """Prepare the json sent to the Predator service for the given test case."""
  result, error_message = _is_predator_testcase(testcase)
  if not result:
    _set_predator_result_with_error(testcase, error_message)
    return None

  crash_revisions_dict, crash_revision_hash = _prepare_component_revisions_dict(
      testcase.crash_revision, testcase.job_type)
  # Do a None check since we can return {} for revision = 0.
  if crash_revisions_dict is None:
    _set_predator_result_with_error(
        testcase, 'Failed to fetch component revisions for revision %s.' %
        testcase.crash_revision)
    return None

  dependency_rolls = []
  start_revision_hash = end_revision_hash = None
  if ':' in testcase.regression:
    regression_parts = testcase.regression.split(':', 1)
    start_revision = int(regression_parts[0])
    end_revision = int(regression_parts[1])

    start_revisions_dict, start_revision_hash = (
        _prepare_component_revisions_dict(start_revision, testcase.job_type))
    # Do a None check since we can return {} for revision = 0.
    if start_revisions_dict is None:
      _set_predator_result_with_error(
          testcase, 'Failed to fetch component revisions for revision %s.' %
          start_revision)
      return None

    end_revisions_dict, end_revision_hash = (
        _prepare_component_revisions_dict(end_revision, testcase.job_type))
    # Do a None check since we can return {} for revision = 0.
    if end_revisions_dict is None:
      _set_predator_result_with_error(
          testcase,
          'Failed to fetch component revisions for revision %s.' % end_revision)
      return None

    if start_revision != 0:
      dependency_rolls = _compute_rolls(start_revisions_dict,
                                        end_revisions_dict)

  # Put the current revisions dictionary in the format predator expects.
  crash_revision_component_revisions_list = (
      _format_component_revisions_for_predator(crash_revisions_dict))

  # In addition to the start and end revisions, Predator expects the regression
  # range to include the dependency path and repository URL in the same way that
  # they would be included in the dependency rolls. Note that we do not take
  # this from the rolls dict directly as it may not be available.
  src_entry = [
      entry for entry in crash_revision_component_revisions_list
      if entry['dep_path'] == 'src'
  ][0]

  # TODO(mbarbella): This is a hack since ClusterFuzz relies on "src" as a
  # special-cased path, but this is only going to be the correct repository
  # root path some of the time. For certain cases, we must update it.
  repo_url = src_entry['repo_url']
  real_dep_path = SRC_COMPONENT_OVERRIDES.get(repo_url, 'src')
  if real_dep_path != 'src':
    for dependency_list in [
        dependency_rolls, crash_revision_component_revisions_list
    ]:
      for entry in dependency_list:
        if entry['dep_path'] == 'src':
          entry['dep_path'] = real_dep_path
          break

  regression_range = {
      'dep_path': real_dep_path,
      'repo_url': repo_url,
      'old_revision': start_revision_hash,
      'new_revision': end_revision_hash,
  }

  crash_stacktrace = _filter_stacktrace(data_handler.get_stacktrace(testcase))

  return pubsub.Message(
      data=json.dumps({
          'stack_trace': crash_stacktrace,
          'crash_revision': crash_revision_hash,
          'customized_data': {
              'regression_range': regression_range,
              'dependency_rolls': dependency_rolls,
              'dependencies': crash_revision_component_revisions_list,
              'crash_type': testcase.crash_type,
              'crash_address': testcase.crash_address,
              'sanitizer': environment.get_memory_tool_name(testcase.job_type),
              'security_flag': testcase.security_flag,
              'job_type': testcase.job_type,
              'testcase_id': testcase.key.id()
          },
          'platform': testcase.platform,
          'client_id': 'clusterfuzz',
          'signature': testcase.crash_state,
      }).encode('utf-8'))


def _clear_blame_result_and_set_pending_flag(testcase):
  """Clear blame result and set pending bit."""
  testcase.set_metadata('blame_pending', True, update_testcase=False)
  testcase.set_metadata('predator_result', None, update_testcase=False)
  testcase.put()


def execute_task(testcase_id, _):
  """Attempt to find the CL introducing the bug associated with testcase_id."""
  # Locate the testcase associated with the id.
  testcase = data_handler.get_testcase_by_id(testcase_id)
  if not testcase:
    return

  # Make sure that predator topic is configured. If not, nothing to do here.
  topic = db_config.get_value('predator_crash_topic')
  if not topic:
    logs.log('Predator is not configured, skipping blame task.')
    return

  data_handler.update_testcase_comment(testcase, data_types.TaskState.STARTED)

  # Prepare pubsub message to send to predator.
  message = _prepare_predator_message(testcase)
  if not message:
    testcase = data_handler.get_testcase_by_id(testcase_id)
    data_handler.update_testcase_comment(
        testcase, data_types.TaskState.ERROR,
        'Failed to generate request for Predator')
    return

  # Clear existing results and mark blame result as pending.
  testcase = data_handler.get_testcase_by_id(testcase_id)
  _clear_blame_result_and_set_pending_flag(testcase)

  # Post request to pub sub.
  client = pubsub.PubSubClient()
  message_ids = client.publish(topic, [message])
  logs.log('Successfully published testcase %s to Predator. Message IDs: %s.' %
           (testcase_id, message_ids))
  data_handler.update_testcase_comment(testcase, data_types.TaskState.FINISHED)



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Stack parser module."""

import inspect

from clusterfuzz._internal.metrics import logs
from clusterfuzz._internal.protos import process_state_pb2


def unsigned_to_signed(address):
  """Convert unsigned address to signed int64 (as defined in the proto)."""
  return (address - 2**64) if address >= 2**63 else address


def format_address_to_dec(address, base=16):
  """Addresses may be formatted as decimal, hex string with 0x or 0X prefix,
     or without any prefix. Convert to decimal int."""
  if address is None:
    return None

  address = str(address).replace('`', '').strip()
  if not address:
    return None

  # This is required for Chrome Win and Mac stacks, which mix decimal and hex.
  try_bases = [base, 16] if base != 16 else [base]
  for base_try in try_bases:
    try:
      address = int(address, base_try)
      return address
    except Exception:
      continue

  logs.log_warn('Error formatting address %s to decimal int64 in bases %s.' %
                (str(address), str(try_bases)))
  return None


class StackFrameStructure(object):
  """IR for fields a stackframe may contain/expect."""

  def __init__(self,
               address=None,
               function_name=None,
               function_base=None,
               function_offset=None,
               filename=None,
               fileline=None,
               module_name=None,
               module_base=None,
               module_offset=None):
    self._address = address
    self._function_name = function_name
    self._function_base = function_base
    self._function_offset = function_offset
    self._filename = filename
    self._fileline = fileline
    self._module_name = module_name
    self._module_base = module_base
    self._module_offset = module_offset

  @property
  def address(self):
    return self._address

  @address.setter
  def address(self, address):
    self._address = address

  @property
  def function_name(self):
    return self._function_name

  @function_name.setter
  def function_name(self, function_name):
    self._function_name = function_name

  @property
  def function_base(self):
    return self._function_base

  @function_base.setter
  def function_base(self, function_base):
    self._function_base = function_base

  @property
  def function_offset(self):
    return self._function_offset

  @function_offset.setter
  def function_offset(self, function_offset):
    self._function_offset = function_offset

  @property
  def filename(self):
    return self._filename

  @filename.setter
  def filename(self, filename):
    self._filename = filename

  @property
  def fileline(self):
    return self._fileline

  @fileline.setter
  def fileline(self, fileline):
    self._fileline = fileline

  @property
  def module_name(self):
    return self._module_name

  @module_name.setter
  def module_name(self, module_name):
    self._module_name = module_name

  @property
  def module_base(self):
    return self._module_base

  @module_base.setter
  def module_base(self, module_base):
    self._module_base = module_base

  @property
  def module_offset(self):
    return self._module_offset

  @module_offset.setter
  def module_offset(self, module_offset):
    self._module_offset = module_offset

  def to_proto(self):
    """Convert StackFrame to process_state.proto format for upload to crash/."""
    frame_proto = process_state_pb2.StackFrame()
    if self.address is not None:
      frame_proto.instruction = unsigned_to_signed(self.address)
    if self.module_base is not None:
      frame_proto.module.base_address = unsigned_to_signed(
          int(self.module_base))
    if self.module_name is not None:
      frame_proto.module.code_file = self.module_name
    if self.function_name is not None:
      frame_proto.function_name = self.function_name
    if self.function_base is not None:
      frame_proto.function_base = unsigned_to_signed(self.function_base)
    if self.filename is not None:
      frame_proto.source_file_name = self.filename
    if self.fileline is not None:
      frame_proto.source_line = int(self.fileline)

    return frame_proto


class StackFrame(StackFrameStructure):
  """IR for canonicalizing stackframe strings."""

  def __init__(self,
               address=None,
               function_name=None,
               function_base=None,
               function_offset=None,
               filename=None,
               fileline=None,
               module_name=None,
               module_base=None,
               module_offset=None,
               base=16):
    super(StackFrame, self).__init__(
        address=format_address_to_dec(address),
        function_name=function_name,
        function_base=format_address_to_dec(function_base),
        function_offset=format_address_to_dec(function_offset),
        filename=filename,
        fileline=fileline,
        module_name=module_name,
        module_base=format_address_to_dec(module_base),
        module_offset=format_address_to_dec(module_offset))

    # Base for converting addresses set in frame. Most will be in hex.
    self._base = base

  def __setattr__(self, field_name, field_value):
    """Set attributes, performing conversions as needed for address fields."""
    if field_name == 'base':
      self._base = field_value
      return

    address_fields = [
        'address',
        'function_base',
        'function_offset',
        'module_base',
        'module_offset',
    ]
    if field_name in address_fields:
      address = format_address_to_dec(field_value, self._base)
      super(StackFrame, self).__setattr__(field_name, address)
      return

    super(StackFrame, self).__setattr__(field_name, field_value)

  def __str__(self):
    s = []
    for name, member in inspect.getmembers(StackFrame):
      if not isinstance(member, property):
        continue
      s += ['%s: %s' % (name, str(getattr(self, name)))]
    return ', '.join(s)


class StackFrameSpec(StackFrameStructure):
  """Representation paralleling that of StackFrames for pulling out the correct
     groups in a *_STACK_FRAME_REGEX match."""

  def __init__(self,
               address=None,
               function_name=None,
               function_base=None,
               function_offset=None,
               filename=None,
               fileline=None,
               module_name=None,
               module_base=None,
               module_offset=None,
               base=16):
    """Specify a stackframe format. Each field should be an index into a match.
       See comments inline *_STACK_FRAME_REGEX for appropriate indices."""
    address = address if address is not None else []
    function_name = function_name if function_name is not None else []
    function_base = function_base if function_base is not None else []
    function_offset = function_offset if function_offset is not None else []
    filename = filename if filename is not None else []
    fileline = fileline if fileline is not None else []
    module_name = module_name if module_name is not None else []
    module_base = module_base if module_base is not None else []
    module_offset = module_offset if module_offset is not None else []
    super(StackFrameSpec, self).__init__(
        address=address,
        function_name=function_name,
        function_base=function_base,
        function_offset=function_offset,
        filename=filename,
        fileline=fileline,
        module_name=module_name,
        module_base=module_base,
        module_offset=module_offset)

    # Base for converting addresses processed by this spec. Most will be in hex.
    self._base = base

  @property
  def base(self):
    return self._base

  @base.setter
  def base(self, base):
    self._base = base

  def parse_stack_frame(self, frame_match):
    """Given a match and stackframe specification, populate a stackframe with
       information from the match."""
    if frame_match is None:
      return None

    frame = StackFrame()

    # Set base to use for address translation.
    frame.base = self.base
    for name, member in inspect.getmembers(StackFrameSpec):
      # Pull out only the property fields; we don't care about methods etc.
      if not isinstance(member, property):
        continue

      # We've already set the base (which we need to do first); skip.
      if name == 'base':
        continue

      # Populate the stackframe field. Try all provided lookup groups.
      indices = getattr(self, name)
      if not isinstance(indices, list):
        indices = [indices]
        setattr(self, name, indices)
      for ind in indices:
        frame_field = frame_match.group(ind)
        if frame_field is not None and frame_field:
          setattr(frame, name, frame_field.strip())
          break

    return frame



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for corpus_pruning_task."""
import os
import shutil
import tempfile
import unittest

# pylint: disable=unused-argument
import mock

from clusterfuzz._internal.base import utils
from clusterfuzz._internal.bot.fuzzers import init as fuzzers_init
from clusterfuzz._internal.bot.tasks import minimize_task
from clusterfuzz._internal.datastore import data_handler
from clusterfuzz._internal.datastore import data_types
from clusterfuzz._internal.google_cloud_utils import blobs
from clusterfuzz._internal.system import environment
from clusterfuzz._internal.tests.test_libs import helpers
from clusterfuzz._internal.tests.test_libs import test_utils
from clusterfuzz._internal.tests.test_libs import untrusted_runner_helpers

TEST_DIR = os.path.join(
    os.path.dirname(os.path.realpath(__file__)), 'minimize_task_data')


@test_utils.with_cloud_emulators('datastore', 'pubsub')
class LibFuzzerMinimizeTaskTest(unittest.TestCase):
  """libFuzzer Minimize task tests."""

  def setUp(self):
    helpers.patch_environ(self)
    helpers.patch(self, [
        'clusterfuzz._internal.bot.tasks.minimize_task._run_libfuzzer_testcase',
        'clusterfuzz._internal.bot.tasks.minimize_task._run_libfuzzer_tool',
    ])

    test_utils.setup_pubsub(utils.get_application_id())

    environment.set_value('APP_ARGS', '%TESTCASE% fuzz_target')
    environment.set_value('APP_DIR', '/libfuzzer')
    environment.set_value('APP_NAME', '')
    environment.set_value('APP_PATH', '')
    environment.set_value('BOT_TMPDIR', '/bot_tmpdir')
    environment.set_value('CRASH_STACKTRACES_DIR', '/crash_stacks')
    environment.set_value('FUZZER_DIR', '/fuzzer_dir')
    environment.set_value('INPUT_DIR', '/input_dir')
    environment.set_value('JOB_NAME', 'libfuzzer_asan_test')
    environment.set_value('USER_PROFILE_IN_MEMORY', True)

  def test_libfuzzer_skip_minimization_initial_crash_state(self):
    """Test libFuzzer minimization skipping with a valid initial crash state."""
    # TODO(ochang): Fix circular import.
    from clusterfuzz._internal.crash_analysis.crash_result import CrashResult

    data_types.Job(name='libfuzzer_asan_job').put()
    testcase = data_types.Testcase(
        minimized_keys='',
        fuzzed_keys='FUZZED_KEY',
        job_type='libfuzzer_asan_job',
        security_flag=True)
    testcase.put()

    stacktrace = (
        '==14970==ERROR: AddressSanitizer: heap-buffer-overflow on address '
        '0x61b00001f7d0 at pc 0x00000064801b bp 0x7ffce478dbd0 sp '
        '0x7ffce478dbc8 READ of size 4 at 0x61b00001f7d0 thread T0\n'
        '#0 0x64801a in frame0() src/test.cpp:1819:15\n'
        '#1 0x647ac5 in frame1() src/test.cpp:1954:25\n'
        '#2 0xb1dee7 in frame2() src/test.cpp:160:9\n'
        '#3 0xb1ddd8 in frame3() src/test.cpp:148:34\n')
    self.mock._run_libfuzzer_testcase.return_value = CrashResult(  # pylint: disable=protected-access
        1, 1.0, stacktrace)

    self.mock._run_libfuzzer_tool.return_value = (None, None)  # pylint: disable=protected-access

    minimize_task.do_libfuzzer_minimization(testcase, '/testcase_file_path')

    testcase = data_handler.get_testcase_by_id(testcase.key.id())
    self.assertEqual('Heap-buffer-overflow', testcase.crash_type)
    self.assertEqual('frame0\nframe1\nframe2\n', testcase.crash_state)
    self.assertEqual('0x61b00001f7d0', testcase.crash_address)
    self.assertEqual(
        '+----------------------------------------Release Build Stacktrace'
        '----------------------------------------+\n%s' % stacktrace,
        testcase.crash_stacktrace)


class MinimizeTaskTestUntrusted(
    untrusted_runner_helpers.UntrustedRunnerIntegrationTest):
  """Minimize task tests for untrusted."""

  def setUp(self):
    """Set up."""
    super().setUp()
    environment.set_value('JOB_NAME', 'libfuzzer_asan_job')

    patcher = mock.patch(
        'clusterfuzz._internal.bot.fuzzers.libFuzzer.fuzzer.LibFuzzer.fuzzer_directory',
        new_callable=mock.PropertyMock)

    mock_fuzzer_directory = patcher.start()
    self.addCleanup(patcher.stop)

    mock_fuzzer_directory.return_value = os.path.join(
        environment.get_value('ROOT_DIR'), 'src', 'clusterfuzz', '_internal',
        'bot', 'fuzzers', 'libFuzzer')

    job = data_types.Job(
        name='libfuzzer_asan_job',
        environment_string=(
            'RELEASE_BUILD_BUCKET_PATH = '
            'gs://clusterfuzz-test-data/test_libfuzzer_builds/'
            'test-libfuzzer-build-([0-9]+).zip\n'
            'REVISION_VARS_URL = https://commondatastorage.googleapis.com/'
            'clusterfuzz-test-data/test_libfuzzer_builds/'
            'test-libfuzzer-build-%s.srcmap.json\n'))
    job.put()

    data_types.FuzzTarget(
        engine='libFuzzer', binary='test_fuzzer', project='test-project').put()
    data_types.FuzzTargetJob(
        fuzz_target_name='libFuzzer_test_fuzzer',
        engine='libFuzzer',
        job='libfuzzer_asan_job').put()

    environment.set_value('USE_MINIJAIL', True)
    data_types.Fuzzer(
        revision=1,
        file_size='builtin',
        source='builtin',
        name='libFuzzer',
        max_testcases=4,
        builtin=True).put()
    self.temp_dir = tempfile.mkdtemp(dir=environment.get_value('FUZZ_INPUTS'))

  def tearDown(self):
    super().tearDown()
    shutil.rmtree(self.temp_dir, ignore_errors=True)

  def test_minimize(self):
    """Test minimize."""
    helpers.patch(self, ['clusterfuzz._internal.base.utils.is_oss_fuzz'])
    self.mock.is_oss_fuzz.return_value = True

    testcase_file_path = os.path.join(self.temp_dir, 'testcase')
    with open(testcase_file_path, 'wb') as f:
      f.write(b'EEE')

    with open(testcase_file_path) as f:
      fuzzed_keys = blobs.write_blob(f)

    testcase_path = os.path.join(self.temp_dir, 'testcase')

    testcase = data_types.Testcase(
        crash_type='Null-dereference WRITE',
        crash_address='',
        crash_state='Foo\n',
        crash_stacktrace='',
        crash_revision=1337,
        fuzzed_keys=fuzzed_keys,
        fuzzer_name='libFuzzer',
        overridden_fuzzer_name='libFuzzer_test_fuzzer',
        job_type='libfuzzer_asan_job',
        absolute_path=testcase_path,
        minimized_arguments='%TESTCASE% test_fuzzer')
    testcase.put()

    data_types.FuzzTarget(engine='libFuzzer', binary='test_fuzzer').put()

    fuzzers_init.run()

    self._setup_env(job_type='libfuzzer_asan_job')
    environment.set_value('APP_ARGS', testcase.minimized_arguments)
    environment.set_value('LIBFUZZER_MINIMIZATION_ROUNDS', 3)
    environment.set_value('UBSAN_OPTIONS',
                          'unneeded_option=1:silence_unsigned_overflow=1')
    minimize_task.execute_task(testcase.key.id(), 'libfuzzer_asan_job')

    testcase = data_handler.get_testcase_by_id(testcase.key.id())
    self.assertNotEqual('', testcase.minimized_keys)
    self.assertNotEqual('NA', testcase.minimized_keys)
    self.assertNotEqual(testcase.fuzzed_keys, testcase.minimized_keys)
    self.assertEqual({
        'ASAN_OPTIONS': {},
        'UBSAN_OPTIONS': {
            'silence_unsigned_overflow': 1
        }
    }, testcase.get_metadata('env'))

    blobs.read_blob_to_disk(testcase.minimized_keys, testcase_path)

    with open(testcase_path, 'rb') as f:
      self.assertEqual(1, len(f.read()))
